# -*- coding: utf-8 -*-

"""Predicates for checking a node's type."""

from typing import Tuple, Type, Union

from .utils import node_predicate
from ..typing import NodePredicate
from ....dsl import (
    Abundance,
    BaseEntity,
    BiologicalProcess,
    CentralDogma,
    ComplexAbundance,
    CompositeAbundance,
    Gene,
    ListAbundance,
    MicroRna,
    Pathology,
    Population,
    Protein,
    Reaction,
    Rna,
    Transcribable,
)

__all__ = [
    "is_abundance",
    "is_biological_process",
    "is_pathology",
    "is_transcribable",
    "is_rna",
    "is_reaction",
    "is_protein",
    "is_population",
    "is_mirna",
    "is_list",
    "is_gene",
    "is_composite",
    "is_central_dogma",
    "is_complex",
    "not_abundance",
    "not_biological_process",
    "not_pathology",
    "not_rna",
    "not_reaction",
    "not_protein",
    "not_population",
    "not_mirna",
    "not_gene",
    "not_composite",
    "not_complex",
    "not_central_dogma",
    "not_list",
    "not_transcribable",
]


def _type_checker(cls: Union[Type[BaseEntity], Tuple[Type[BaseEntity], ...]]) -> NodePredicate:
    @node_predicate
    def _is_type(node: BaseEntity) -> bool:
        return isinstance(node, cls)

    return _is_type


def _not_type_checker(cls: Union[Type[BaseEntity], Tuple[Type[BaseEntity], ...]]) -> NodePredicate:
    @node_predicate
    def _not_type(node: BaseEntity) -> bool:
        return not isinstance(node, cls)

    return _not_type


is_abundance = _type_checker(Abundance)
not_abundance = _not_type_checker(Abundance)

is_biological_process = _type_checker(BiologicalProcess)
not_biological_process = _not_type_checker(BiologicalProcess)

is_pathology = _type_checker(Pathology)
not_pathology = _not_type_checker(Pathology)

is_population = _type_checker(Population)
not_population = _not_type_checker(Population)

#: Return true if the node is a gene, RNA, miRNA, or Protein
is_central_dogma = _type_checker(CentralDogma)
not_central_dogma = _not_type_checker(CentralDogma)

is_gene = _type_checker(Gene)
not_gene = _not_type_checker(Gene)

is_transcribable = _type_checker(Transcribable)
not_transcribable = _not_type_checker(Transcribable)

is_rna = _type_checker(Rna)
not_rna = _not_type_checker(Rna)

is_mirna = _type_checker(MicroRna)
not_mirna = _not_type_checker(MicroRna)

is_protein = _type_checker(Protein)
not_protein = _not_type_checker(Protein)

is_list = _type_checker(ListAbundance)
not_list = _not_type_checker(ListAbundance)

is_composite = _type_checker(CompositeAbundance)
not_composite = _not_type_checker(CompositeAbundance)

is_complex = _type_checker(ComplexAbundance)
not_complex = _not_type_checker(ComplexAbundance)

is_reaction = _type_checker(Reaction)
not_reaction = _not_type_checker(Reaction)



# -*- coding: utf-8 -*-

"""Tests for node deletion functions."""

import unittest

from pybel import BELGraph
from pybel.constants import POSITIVE_CORRELATION, RELATION
from pybel.dsl import (
    CompositeAbundance,
    Protein,
    gene,
    hgvs,
    pathology,
    protein_fusion,
    rna,
    rna_fusion,
)
from pybel.struct.mutation import (
    enrich_protein_and_rna_origins,
    prune_protein_rna_origins,
    remove_associations,
    remove_isolated_list_abundances,
    remove_pathologies,
)
from pybel.struct.mutation.utils import remove_isolated_nodes, remove_isolated_nodes_op
from pybel.testing.utils import n

trem2_gene = gene(namespace="HGNC", name="TREM2")
trem2_rna = rna(namespace="HGNC", name="TREM2")
trem2_protein = Protein(namespace="HGNC", name="TREM2")


class TestDeletions(unittest.TestCase):
    """Test cases for deletion functions."""

    def test_remove_pathologies(self):
        """Test removal of pathologies."""
        g = BELGraph()

        p1, p2, p3 = (Protein(namespace="HGNC", name=n()) for _ in range(3))
        d1, d2 = (pathology(namespace="MESH", name=n()) for _ in range(2))

        g.add_increases(p1, p2, citation=n(), evidence=n())
        g.add_increases(p2, p3, citation=n(), evidence=n())
        g.add_positive_correlation(p1, d1, citation=n(), evidence=n())
        g.add_positive_correlation(p2, d1, citation=n(), evidence=n())
        g.add_association(p2, d1, citation=n(), evidence=n())
        g.add_positive_correlation(d1, d2, citation=n(), evidence=n())
        g.add_positive_correlation(d1, d2, citation=n(), evidence=n())

        self.assertEqual(5, g.number_of_nodes())
        self.assertEqual(12, g.number_of_edges())
        self.assertEqual(2, len(g[p2][d1]))

        remove_associations(g)

        relations = list(g[p2][d1].values())
        self.assertEqual(1, len(relations))
        self.assertEqual(POSITIVE_CORRELATION, relations[0][RELATION])

        self.assertEqual(5, g.number_of_nodes())
        self.assertEqual(10, g.number_of_edges())
        self.assertEqual(5, g.number_of_nodes())

        remove_pathologies(g)

        self.assertTrue(p1, g)
        self.assertTrue(p2, g)
        self.assertTrue(p3, g)
        self.assertEqual(3, g.number_of_nodes())
        self.assertEqual(2, g.number_of_edges())

    def test_remove_isolated_in_place(self):
        """Test removing isolated nodes (in-place)."""
        g = BELGraph()

        g.add_edge(1, 2)
        g.add_edge(2, 3)
        g.add_node(4)

        remove_isolated_nodes(g)

        self.assertEqual(3, g.number_of_nodes())
        self.assertEqual(2, g.number_of_edges())

    def test_remove_isolated_out_of_place(self):
        """Test removing isolated nodes (out-of-place)."""
        g = BELGraph()

        g.add_edge(1, 2)
        g.add_edge(2, 3)
        g.add_node(4)

        g = remove_isolated_nodes_op(g)

        self.assertEqual(3, g.number_of_nodes())
        self.assertEqual(2, g.number_of_edges())

    def test_remove_isolated_list_abundances(self):
        """Test removing isolated list abundances."""
        g = BELGraph()
        p1, p2 = [Protein("HGNC", n()) for _ in range(2)]
        node = CompositeAbundance([p1, p2])
        g.add_node_from_data(node)
        self.assertEqual(3, g.number_of_nodes())
        remove_isolated_list_abundances(g)
        self.assertEqual(2, g.number_of_nodes())
        self.assertIn(p1, g)
        self.assertIn(p2, g)


class TestProcessing(unittest.TestCase):
    """Test inference of the central dogma."""

    def assert_in_graph(self, node, graph):
        """Assert the node is in the graph.

        :type node: pybel.dsl.BaseEntity
        :type graph: pybel.BELGraph
        :rtype: bool
        """
        self.assertIn(node, graph)

    def assert_not_in_graph(self, node, graph):
        """Assert the node is not in the graph.

        :type node: pybel.dsl.BaseEntity
        :type graph: pybel.BELGraph
        :rtype: bool
        """
        self.assertNotIn(node, graph)

    def test_infer_on_sialic_acid_example(self):
        """Test infer_central_dogma on the sialic acid example."""
        graph = BELGraph()
        graph.add_node_from_data(trem2_protein)

        self.assert_in_graph(trem2_protein, graph)
        self.assert_not_in_graph(trem2_gene, graph)
        self.assert_not_in_graph(trem2_rna, graph)

        enrich_protein_and_rna_origins(graph)

        self.assert_in_graph(trem2_gene, graph)
        self.assert_in_graph(trem2_rna, graph)

        prune_protein_rna_origins(graph)

        self.assert_not_in_graph(trem2_gene, graph)
        self.assert_not_in_graph(trem2_rna, graph)
        self.assert_in_graph(trem2_protein, graph)

    def test_no_infer_on_protein_variants(self):
        """Test that expansion doesn't occur on protein variants."""
        p = Protein("HGNC", n(), variants=[hgvs(n())])

        graph = BELGraph()
        graph.add_node_from_data(p)

        self.assertEqual(2, graph.number_of_nodes())
        self.assertEqual(1, graph.number_of_edges())

        enrich_protein_and_rna_origins(graph)

        self.assertEqual(4, graph.number_of_nodes())
        self.assertEqual(3, graph.number_of_edges())

    def test_no_infer_on_rna_variants(self):
        """Test that expansion doesn't occur on RNA variants."""
        r = rna("HGNC", n(), variants=[hgvs(n())])

        graph = BELGraph()
        graph.add_node_from_data(r)

        self.assertEqual(2, graph.number_of_nodes())
        self.assertEqual(1, graph.number_of_edges())

        enrich_protein_and_rna_origins(graph)

        self.assertEqual(3, graph.number_of_nodes())
        self.assertEqual(2, graph.number_of_edges())

    def test_no_infer_protein_fusion(self):
        """Test that no gene is inferred from a RNA fusion node."""
        partner5p = Protein(n(), n())
        partner3p = Protein(n(), n())

        p = protein_fusion(partner_3p=partner3p, partner_5p=partner5p)

        graph = BELGraph()
        graph.add_node_from_data(p)

        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())

        enrich_protein_and_rna_origins(graph)

        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())

    def test_no_infer_rna_fusion(self):
        """Test that no RNA nor gene is inferred from a protein fusion node."""
        partner5p = rna(n(), n())
        partner3p = rna(n(), n())

        p = rna_fusion(partner_3p=partner3p, partner_5p=partner5p)

        graph = BELGraph()
        graph.add_node_from_data(p)

        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())

        enrich_protein_and_rna_origins(graph)

        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())



# -*- coding: utf-8 -*-

"""Fragments.

The addition of a fragment results in an entry called :data:`pybel.constants.VARIANTS`
in the data dictionary associated with a given node. This entry is a list with dictionaries
describing each of the variants. All variants have the entry :data:`pybel.constants.KIND` to identify whether it is
a PTM, gene modification, fragment, or HGVS variant. The :data:`pybel.constants.KIND` value for a fragment is
:data:`pybel.constants.FRAGMENT`.

Each fragment contains an identifier, which is a dictionary with the namespace and name, and can optionally include
the position ('pos') and/or amino acid code ('code').

For example, the node :code:`p(HGNC:GSK3B, frag(45_129))` is represented with the following:

.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: PROTEIN,
        NAMESPACE: 'HGNC',
        NAME: 'GSK3B',
        VARIANTS: [
            {
                KIND: FRAGMENT,
                FRAGMENT_START: 45,
                FRAGMENT_STOP: 129,
            },
        ],
    }

Additionally, nodes can have an asterick (*) or question mark (?) representing unbound
or unknown fragments, respectively.

A fragment may also be unknown, such as in the node :code:`p(HGNC:GSK3B, frag(?))`. This
is represented with the key :data:`pybel.constants.FRAGMENT_MISSING` and the value of '?' like:


.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: PROTEIN,
        NAMESPACE: 'HGNC',
        NAME: 'GSK3B',
        VARIANTS: [
            {
                KIND: FRAGMENT,
                FRAGMENT_MISSING: '?',
            },
        ],
    }

.. seealso::

    - BEL 2.0 specification on `proteolytic fragments (2.2.3) <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_proteolytic_fragments>`_
    - PyBEL module :py:class:`pybel.parser.modifiers.get_fragment_language`
"""

from pyparsing import And, Keyword, Optional, ParserElement, Suppress
from pyparsing import pyparsing_common as ppc

from ..utils import WCW, nest, one_of_tags, quote
from ...constants import (
    FRAGMENT,
    FRAGMENT_DESCRIPTION,
    FRAGMENT_MISSING,
    FRAGMENT_START,
    FRAGMENT_STOP,
    KIND,
)

__all__ = [
    "get_fragment_language",
]

fragment_tag = one_of_tags(tags=["frag", "fragment"], canonical_tag=FRAGMENT, name=KIND)
fragment_range = (ppc.integer | "?")(FRAGMENT_START) + "_" + (ppc.integer | "?" | "*")(FRAGMENT_STOP)
missing_fragment = Keyword("?")(FRAGMENT_MISSING)


def get_fragment_language() -> ParserElement:
    """Build a protein fragment parser."""
    _fragment_value_inner = fragment_range | missing_fragment(FRAGMENT_MISSING)
    _fragment_value = _fragment_value_inner | And([Suppress('"'), _fragment_value_inner, Suppress('"')])
    parser_element = fragment_tag + nest(_fragment_value + Optional(WCW + quote(FRAGMENT_DESCRIPTION)))
    return parser_element



# -*- coding: utf-8 -*-

"""Utilities for functions for collapsing nodes."""

import itertools as itt
from typing import Mapping, Set

from ...filters.edge_filters import filter_edges
from ...filters.edge_predicate_builders import build_relation_predicate
from ...pipeline import in_place_transformation
from ....constants import HAS_VARIANT
from ....dsl import BaseEntity
from ....utils import hash_edge

__all__ = [
    "collapse_pair",
    "collapse_nodes",
    "collapse_all_variants",
    "surviors_are_inconsistent",
]


def _remove_self_edges(graph):
    self_edges = [(u, u, k) for u in graph if u in graph[u] for k in graph[u][u]]
    graph.remove_edges_from(self_edges)


@in_place_transformation
def collapse_pair(graph, survivor: BaseEntity, victim: BaseEntity) -> None:
    """Rewire all edges from the synonymous node to the survivor node, then deletes the synonymous node.

    Does not keep edges between the two nodes.

    :param pybel.BELGraph graph: A BEL graph
    :param survivor: The BEL node to collapse all edges on the synonym to
    :param victim: The BEL node to collapse into the surviving node
    """
    graph.add_edges_from(
        (survivor, successor, hash_edge(survivor, successor, edge_data), edge_data)
        for _, successor, edge_data in graph.out_edges(victim, data=True)
        if successor != survivor
    )

    graph.add_edges_from(
        (predecessor, survivor, hash_edge(predecessor, survivor, edge_data), edge_data)
        for predecessor, _, edge_data in graph.in_edges(victim, data=True)
        if predecessor != survivor
    )

    if victim in graph:
        graph.remove_node(victim)


# TODO what happens when collapsing is not consistent? Need to build intermediate mappings and test their consistency.


@in_place_transformation
def collapse_nodes(graph, survivor_mapping: Mapping[BaseEntity, Set[BaseEntity]]) -> None:
    """Collapse all nodes in values to the key nodes, in place.

    :param pybel.BELGraph graph: A BEL graph
    :param survivor_mapping: A dictionary with survivors as their keys, and iterables of the corresponding victims as
     values.
    """
    inconsistencies = surviors_are_inconsistent(survivor_mapping)
    if inconsistencies:
        raise ValueError("survivor mapping is inconsistent: {}".format(inconsistencies))

    for survivor, victims in survivor_mapping.items():
        for victim in victims:
            collapse_pair(graph, survivor=survivor, victim=victim)

    _remove_self_edges(graph)


def surviors_are_inconsistent(survivor_mapping: Mapping[BaseEntity, Set[BaseEntity]]) -> Set[BaseEntity]:
    """Check that there's no transitive shit going on."""
    victim_mapping = set()
    for victim in itt.chain.from_iterable(survivor_mapping.values()):
        if victim in survivor_mapping:
            victim_mapping.add(victim)
    return victim_mapping


@in_place_transformation
def collapse_all_variants(graph) -> None:
    """Collapse all genes', RNAs', miRNAs', and proteins' variants to their parents.

    :param pybel.BELGraph graph: A BEL Graph
    """
    has_variant_predicate = build_relation_predicate(HAS_VARIANT)

    edges = list(filter_edges(graph, has_variant_predicate))

    for u, v, _ in edges:
        collapse_pair(graph, survivor=u, victim=v)

    _remove_self_edges(graph)



# -*- coding: utf-8 -*-

"""Tests for the query builder."""

import logging
import unittest

from pybel import BELGraph, Pipeline
from pybel.dsl import Protein
from pybel.examples.egf_example import egf_graph, vcp
from pybel.examples.homology_example import (
    homology_graph,
    mouse_csf1_protein,
    mouse_csf1_rna,
    mouse_mapk1_protein,
    mouse_mapk1_rna,
)
from pybel.examples.sialic_acid_example import (
    cd33_phosphorylated,
    dap12,
    shp1,
    shp2,
    sialic_acid_graph,
    syk,
    trem2,
)
from pybel.struct import (
    expand_node_neighborhood,
    expand_nodes_neighborhoods,
    get_subgraph_by_annotation_value,
)
from pybel.struct.mutation import collapse_to_genes, enrich_protein_and_rna_origins
from pybel.struct.query import Query, QueryMissingNetworksError, Seeding
from pybel.testing.generate import generate_random_graph
from pybel.testing.mock_manager import MockQueryManager
from pybel.testing.utils import n

log = logging.getLogger(__name__)


def add(query, manager, graph):
    network = manager.insert_graph(graph)
    query.append_network(network.id)


class TestSeedingConstructor(unittest.TestCase):
    def test_none(self):
        """Test construction of a seeding container."""
        seeding = Seeding()
        self.assertEqual(0, len(seeding))
        self.assertEqual("[]", seeding.dumps())

    def test_append_sample(self):
        seeding = Seeding()
        seeding.append_sample()
        self.assertEqual(1, len(seeding))

        s = seeding.dumps()
        self.assertIsInstance(s, str)


class TestQueryConstructor(unittest.TestCase):
    """Test the construction of a Query."""

    def test_network_ids_none(self):
        query = Query()
        self.assertIsInstance(query.network_ids, list)
        self.assertIsInstance(query.seeding, Seeding)
        self.assertIsInstance(query.pipeline, Pipeline)
        self.assertEqual(0, len(query.network_ids))

    def test_network_ids_single(self):
        query = Query(network_ids=1)
        self.assertIsInstance(query.network_ids, list)
        self.assertEqual(1, len(query.network_ids))

    def test_network_ids_multiple(self):
        query = Query(network_ids=[1, 2, 3])
        self.assertIsInstance(query.network_ids, list)
        self.assertEqual(3, len(query.network_ids))

    def test_network_ids_type_error(self):
        with self.assertRaises(TypeError):
            Query(network_ids="a")

    def test_seeding(self):
        query = Query(seeding=Seeding())
        self.assertEqual(0, len(query.seeding))

    def test_pipeline(self):
        query = Query(pipeline=Pipeline())
        self.assertEqual(0, len(query.pipeline))


class QueryTestEgf(unittest.TestCase):
    """Test querying the EGF subgraph"""

    def setUp(self):
        """Set up each test with a mock query manager."""
        self.manager = MockQueryManager()
        self.query = Query()

    def add_query(self, graph):
        add(self.query, self.manager, graph)
        return self.query

    def run_query(self):
        return self.query.run(self.manager)

    def test_fail_run_with_no_networks(self):
        with self.assertRaises(QueryMissingNetworksError):
            self.run_query()

    def test_no_seeding_no_pipeline(self):
        graph = egf_graph.copy()

        self.add_query(graph)
        result = self.run_query()

        self.assertEqual(graph.number_of_nodes(), result.number_of_nodes())
        self.assertEqual(graph.number_of_edges(), result.number_of_edges())

    def test_seed_by_neighbor(self):
        graph = BELGraph()
        a, b, c, d = (Protein(namespace=n(), name=str(i)) for i in range(4))

        graph.add_increases(a, b, citation=n(), evidence=n())
        graph.add_increases(b, c, citation=n(), evidence=n())
        graph.add_increases(c, d, citation=n(), evidence=n())

        self.add_query(graph).append_seeding_neighbors(b)
        result = self.run_query()
        self.assertIsInstance(result, BELGraph)
        # test nodes
        self.assertIn(a, result)
        self.assertIn(b, result)
        self.assertIn(c, result)
        self.assertNotIn(d, result)
        # test edges
        self.assertIn(b, result[a])
        self.assertIn(c, result[b])
        self.assertNotIn(d, result[c])

    def test_seed_by_neighbors(self):
        graph = BELGraph()
        a, b, c, d, e = (Protein(namespace=n(), name=str(i)) for i in range(5))

        graph.add_increases(a, b, citation=n(), evidence=n())
        graph.add_increases(b, c, citation=n(), evidence=n())
        graph.add_increases(c, d, citation=n(), evidence=n())
        graph.add_increases(d, e, citation=n(), evidence=n())

        self.add_query(graph).append_seeding_neighbors([b, c])

        result = self.run_query()
        self.assertIsInstance(result, BELGraph)
        # test nodes
        self.assertIn(a, result)
        self.assertIn(b, result)
        self.assertIn(c, result)
        self.assertIn(d, result)
        self.assertNotIn(e, result)
        # test edges
        self.assertIn(b, result[a])
        self.assertIn(c, result[b])
        self.assertIn(d, result[c])
        self.assertNotIn(e, result[d])

    def test_random_sample(self):
        """Test generating multiple random samples and combining them."""
        graph = generate_random_graph(50, 1000)

        query = self.add_query(graph)
        query.append_seeding_sample(number_edges=10)
        query.append_seeding_sample(number_edges=10)

        result = self.run_query()
        self.assertIn(
            result.number_of_edges(),
            {16, 17, 18, 19, 20},
            msg="This will rail randomly sometimes, lol",
        )


class QueryTest(unittest.TestCase):
    """Test the query"""

    def setUp(self):
        """Setup each test with an empty mock query manager."""
        self.manager = MockQueryManager()

    def test_pipeline(self):
        graph = egf_graph.copy()
        enrich_protein_and_rna_origins(graph)

        self.assertEqual(
            32,  # 10 protein nodes already there + complex + bp +  2*10 (genes and rnas)
            graph.number_of_nodes(),
        )

        # 6 already there + 5 complex hasComponent edges + new 2*10 edges
        self.assertEqual(31, graph.number_of_edges())

        network = self.manager.insert_graph(graph)

        pipeline = Pipeline()
        pipeline.append(collapse_to_genes)

        query = Query(network_ids=[network.id], pipeline=pipeline)
        result_graph = query.run(self.manager)

        self.assertEqual(12, result_graph.number_of_nodes())  # same number of nodes than there were
        self.assertEqual(11, result_graph.number_of_edges())  # same number of edges than there were

    def test_pipeline_2(self):
        graph = egf_graph.copy()

        network = self.manager.insert_graph(graph)
        network_id = network.id

        query = Query(network_ids=[network_id])
        query.append_seeding_neighbors(vcp)
        query.append_pipeline(get_subgraph_by_annotation_value, "Species", "9606")

        result = query.run(self.manager)
        self.assertIsNotNone(result, msg="Query returned none")

        self.assertEqual(3, result.number_of_nodes())

    def test_query_multiple_networks(self):
        sialic_acid_graph_id = self.manager.insert_graph(sialic_acid_graph.copy()).id
        egf_graph_id = self.manager.insert_graph(egf_graph.copy()).id

        query = Query()
        query.append_network(sialic_acid_graph_id)
        query.append_network(egf_graph_id)
        query.append_seeding_neighbors([syk])
        query.append_pipeline(enrich_protein_and_rna_origins)

        result = query.run(self.manager)
        self.assertIsNotNone(result, msg="Query returned none")

        self.assertIn(shp1, result)
        self.assertIn(shp2, result)
        self.assertIn(trem2, result)
        self.assertIn(dap12, result)

        self.assertEqual(15, result.number_of_nodes())
        self.assertEqual(14, result.number_of_edges())

    def test_get_subgraph_by_annotation_value(self):
        graph = homology_graph.copy()

        result = get_subgraph_by_annotation_value(graph, "Species", "10090")

        self.assertIsNotNone(result, msg="Query returned none")
        self.assertIsInstance(result, BELGraph)
        self.assertLess(0, result.number_of_nodes())

        self.assertIn(mouse_mapk1_protein, result, msg="nodes:\n{}".format(list(map(repr, graph))))
        self.assertIn(mouse_csf1_protein, result)

        self.assertEqual(2, result.number_of_nodes())
        self.assertEqual(1, result.number_of_edges())

    def test_seeding_1(self):
        test_network_1 = self.manager.insert_graph(homology_graph.copy())

        query = Query(network_ids=[test_network_1.id])
        query.append_seeding_neighbors([mouse_csf1_rna, mouse_mapk1_rna])

        result = query.run(self.manager)
        self.assertIsNotNone(result, msg="Query returned none")
        self.assertIsInstance(result, BELGraph)

        self.assertIn(mouse_mapk1_rna, result)
        self.assertIn(mouse_csf1_rna, result)
        self.assertIn(mouse_mapk1_protein, result)
        self.assertIn(mouse_csf1_protein, result)

        self.assertEqual(6, result.number_of_nodes())
        self.assertEqual(4, result.number_of_edges())

    def test_seeding_with_pipeline(self):
        test_network_1 = self.manager.insert_graph(sialic_acid_graph.copy())

        query = Query(network_ids=[test_network_1.id])
        query.append_seeding_neighbors([trem2, dap12, shp2])
        query.append_pipeline(expand_nodes_neighborhoods, [trem2, dap12, shp2])
        result = query.run(self.manager)
        self.assertIsNotNone(result, msg="Query returned none")
        self.assertIsInstance(result, BELGraph)

        self.assertIn(trem2, result)
        self.assertIn(dap12, result)
        self.assertIn(shp2, result)
        self.assertIn(syk, result)
        self.assertIn(cd33_phosphorylated, result)

        self.assertEqual(5, result.number_of_nodes())
        self.assertEqual(4, result.number_of_edges())

    def test_query_multiple_networks_with_api(self):
        test_network_1 = self.manager.insert_graph(homology_graph.copy())

        pipeline = Pipeline()
        pipeline.append(expand_node_neighborhood, mouse_mapk1_protein)

        query = Query(network_ids=[test_network_1.id], pipeline=pipeline)
        query.append_seeding_annotation("Species", {"10090"})

        result = query.run(self.manager)

        self.assertIsNotNone(result, msg="Query returned none")

        self.assertEqual(3, result.number_of_nodes())
        self.assertIn(mouse_mapk1_protein, result)
        self.assertIn(mouse_csf1_protein, result)

        self.assertEqual(2, result.number_of_edges())



# -*- coding: utf-8 -*-

"""This module contains tests for the SPIA exporter."""

import unittest

from pandas import DataFrame

from pybel.dsl import activity, composite_abundance, pmod, protein, rna
from pybel.examples.sialic_acid_example import (
    cd33,
    citation,
    evidence_1,
    shp1,
    shp2,
    sialic_acid_cd33_complex,
    sialic_acid_graph,
    trem2,
)
from pybel.io.spia import (
    build_spia_matrices,
    get_matrix_index,
    to_spia_dfs,
    update_spia_matrices,
)


class TestSpia(unittest.TestCase):
    """Test SPIA Exporter."""

    def setUp(self):
        self.sialic_acid_graph = sialic_acid_graph.copy()

    def test_build_matrix(self):
        """Test build empty matrix."""

        node_names = get_matrix_index(self.sialic_acid_graph)

        matrix_dict = build_spia_matrices(node_names)

        nodes = {"PTPN11", "TREM2", "PTPN6", "TYROBP", "CD33", "SYK"}

        self.assertEqual(set(matrix_dict["activation"].columns), nodes)
        self.assertEqual(set(matrix_dict["repression"].index), nodes)

    def test_update_matrix_inhibition_ubiquination(self):
        """Test updating the matrix with an inhibition ubiquitination."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2", variants=[pmod("Ub")])

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        # Initialize matrix correctly
        self.assertEqual(test_matrix.values.all(), 0)

        test_dict["inhibition_ubiquination"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "decreases"})

        self.assertEqual(test_dict["inhibition_ubiquination"]["A"]["B"], 1)
        self.assertEqual(test_dict["inhibition_ubiquination"]["A"]["A"], 0)
        self.assertEqual(test_dict["inhibition_ubiquination"]["B"]["A"], 0)
        self.assertEqual(test_dict["inhibition_ubiquination"]["B"]["B"], 0)

    def test_update_matrix_activation_ubiquination(self):
        """Test updating the matrix with an activation ubiquitination."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2", variants=[pmod("Ub")])

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["activation_ubiquination"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "increases"})

        self.assertEqual(test_dict["activation_ubiquination"]["A"]["B"], 1)
        self.assertEqual(test_dict["activation_ubiquination"]["A"]["A"], 0)
        self.assertEqual(test_dict["activation_ubiquination"]["B"]["A"], 0)
        self.assertEqual(test_dict["activation_ubiquination"]["B"]["B"], 0)

    def test_update_matrix_inhibition_phosphorylation(self):
        """Test updating the matrix with an inhibition phosphorylation."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2", variants=[pmod("Ph")])

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["inhibition_phosphorylation"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "decreases"})

        self.assertEqual(test_dict["inhibition_phosphorylation"]["A"]["B"], 1)
        self.assertEqual(test_dict["inhibition_phosphorylation"]["A"]["A"], 0)
        self.assertEqual(test_dict["inhibition_phosphorylation"]["B"]["A"], 0)
        self.assertEqual(test_dict["inhibition_phosphorylation"]["B"]["B"], 0)

    def test_update_matrix_activation_phosphorylation(self):
        """Test updating the matrix with an activation phosphorylation."""

        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2", variants=[pmod("Ph")])

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["activation_phosphorylation"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "increases"})

        self.assertEqual(test_dict["activation_phosphorylation"]["A"]["B"], 1)
        self.assertEqual(test_dict["activation_phosphorylation"]["A"]["A"], 0)
        self.assertEqual(test_dict["activation_phosphorylation"]["B"]["A"], 0)
        self.assertEqual(test_dict["activation_phosphorylation"]["B"]["B"], 0)

    def test_update_matrix_expression(self):
        """Test updating the matrix with RNA expression."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = rna(namespace="HGNC", name="B", identifier="2")

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["expression"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "increases"})

        self.assertEqual(test_dict["expression"]["A"]["B"], 1)
        self.assertEqual(test_dict["expression"]["A"]["A"], 0)
        self.assertEqual(test_dict["expression"]["B"]["A"], 0)
        self.assertEqual(test_dict["expression"]["B"]["B"], 0)

    def test_update_matrix_repression(self):
        """Test updating the matrix with RNA repression."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = rna(namespace="HGNC", name="B", identifier="2")

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["repression"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "decreases"})

        self.assertEqual(test_dict["repression"]["A"]["B"], 1)
        self.assertEqual(test_dict["repression"]["A"]["A"], 0)
        self.assertEqual(test_dict["repression"]["B"]["A"], 0)
        self.assertEqual(test_dict["repression"]["B"]["B"], 0)

    def test_update_matrix_activation(self):
        """Test updating the matrix with activation."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2")

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["activation"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "increases"})

        self.assertEqual(test_dict["activation"]["A"]["B"], 1)
        self.assertEqual(test_dict["activation"]["A"]["A"], 0)
        self.assertEqual(test_dict["activation"]["B"]["A"], 0)
        self.assertEqual(test_dict["activation"]["B"]["B"], 0)

    def test_update_matrix_inhibition(self):
        """Test updating the matrix with activation."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2")

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["inhibition"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "decreases"})

        self.assertEqual(test_dict["inhibition"]["A"]["B"], 1)
        self.assertEqual(test_dict["inhibition"]["A"]["A"], 0)
        self.assertEqual(test_dict["inhibition"]["B"]["A"], 0)
        self.assertEqual(test_dict["inhibition"]["B"]["B"], 0)

    def test_update_matrix_association(self):
        """Test updating the matrix with association."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(namespace="HGNC", name="B", identifier="2")

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["binding_association"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "association"})

        self.assertEqual(test_dict["binding_association"]["A"]["B"], 1)
        self.assertEqual(test_dict["binding_association"]["A"]["A"], 0)
        self.assertEqual(test_dict["binding_association"]["B"]["A"], 0)
        self.assertEqual(test_dict["binding_association"]["B"]["B"], 0)

    def test_update_matrix_pmods(self):
        """Test updating the matrix with multiple protein modifications."""
        sub = protein(namespace="HGNC", name="A", identifier="1")
        obj = protein(
            namespace="HGNC",
            name="B",
            identifier="2",
            variants=[pmod("Ub"), pmod("Ph")],
        )

        index = {"A", "B"}

        test_dict = {}

        test_matrix = DataFrame(0, index=index, columns=index)

        test_dict["activation_ubiquination"] = test_matrix
        test_dict["activation_phosphorylation"] = test_matrix

        update_spia_matrices(test_dict, sub, obj, {"relation": "increases"})

        self.assertEqual(test_dict["activation_ubiquination"]["A"]["B"], 1)
        self.assertEqual(test_dict["activation_ubiquination"]["A"]["A"], 0)
        self.assertEqual(test_dict["activation_ubiquination"]["B"]["A"], 0)
        self.assertEqual(test_dict["activation_ubiquination"]["B"]["B"], 0)

        self.assertEqual(test_dict["activation_phosphorylation"]["A"]["B"], 1)
        self.assertEqual(test_dict["activation_phosphorylation"]["A"]["A"], 0)
        self.assertEqual(test_dict["activation_phosphorylation"]["B"]["A"], 0)
        self.assertEqual(test_dict["activation_phosphorylation"]["B"]["B"], 0)

    def test_spia_matrix_complexes(self):
        """Test handling of complexes."""
        self.sialic_acid_graph.add_increases(
            sialic_acid_cd33_complex,
            trem2,
            citation=citation,
            annotations={"Species": "9606", "Confidence": "High"},
            evidence=evidence_1,
            target_modifier=activity(),
        )

        spia_dfs = to_spia_dfs(self.sialic_acid_graph)

        self.assertEqual(spia_dfs["activation"][cd33.name][trem2.name], 1)

    def test_spia_matrix_composites(self):
        """Test handling of composites."""
        shp = composite_abundance([shp1, shp2])

        self.sialic_acid_graph.add_increases(
            shp,
            trem2,
            citation=citation,
            annotations={"Species": "9606", "Confidence": "High"},
            evidence=evidence_1,
            target_modifier=activity(),
        )

        spia_dfs = to_spia_dfs(self.sialic_acid_graph)

        self.assertEqual(spia_dfs["activation"][shp1.name][trem2.name], 1)
        self.assertEqual(spia_dfs["activation"][shp2.name][trem2.name], 1)



# -*- coding: utf-8 -*-

import os
from pathlib import Path

from pybel import BELGraph, Manager
from pybel.constants import ANNOTATIONS
from pybel.resources import HGNC_URL
from pybel.testing.cases import TemporaryCacheClsMixin
from pybel.testing.constants import belns_dir_path
from pybel.testing.mocks import mock_bel_resources
from tests.constants import OPENBEL_ANNOTATION_RESOURCES

ns1 = Path(os.path.join(belns_dir_path, "disease-ontology.belns")).as_uri()
ns1_url = "http://resources.openbel.org/belframework/20150611/namespace/disease-ontology-ids.belns"

ns2 = Path(os.path.join(belns_dir_path, "mesh-diseases.belns")).as_uri()
ns2_url = "http://resources.openbel.org/belframework/20150611/namespace/mesh-diseases.belns"

CELL_LINE_URL = OPENBEL_ANNOTATION_RESOURCES + "cell-line.belanno"
CELL_LINE_KEYWORD = "CellLine"


class TestDefinitionManagers(TemporaryCacheClsMixin):
    def _help_check_hgnc(self, manager: Manager) -> None:
        """Help check the HGNC namespace was loaded properly."""
        entry = manager.get_namespace_entry(HGNC_URL, "MHS2")
        self.assertIsNotNone(entry)
        self.assertEqual("MHS2", entry.name)
        self.assertIn("G", entry.encoding)

        entry = manager.get_namespace_entry(HGNC_URL, "MIATNB")
        self.assertIsNotNone(entry)
        self.assertEqual("MIATNB", entry.name)
        self.assertIn("G", entry.encoding)
        self.assertIn("R", entry.encoding)

        entry = manager.get_namespace_entry(HGNC_URL, "MIA")
        self.assertIsNotNone(entry)
        self.assertEqual("MIA", entry.name)
        self.assertIn("G", entry.encoding)
        self.assertIn("P", entry.encoding)
        self.assertIn("R", entry.encoding)

    @mock_bel_resources
    def test_insert_namespace_persistent(self, mock_get):
        self.assertEqual(0, self.manager.count_namespaces())
        self.assertEqual(0, self.manager.count_namespace_entries())
        self.manager.get_or_create_namespace(HGNC_URL)
        self._help_check_hgnc(self.manager)

        self.manager.get_or_create_namespace(HGNC_URL)
        self._help_check_hgnc(self.manager)

        self.manager.drop_namespace_by_url(HGNC_URL)
        self.assertEqual(0, self.manager.count_namespaces())
        self.assertEqual(0, self.manager.count_namespace_entries())

    @mock_bel_resources
    def test_insert_annotation(self, mock_get):
        self.assertEqual(0, self.manager.count_annotations())
        self.assertEqual(0, self.manager.count_annotation_entries())
        annotation = self.manager.get_or_create_annotation(CELL_LINE_URL)
        self.assertIsNotNone(annotation)
        self.assertEqual(CELL_LINE_URL, annotation.url)

        entry = self.manager.get_namespace_entry(CELL_LINE_URL, "1321N1 cell")
        self.assertEqual("1321N1 cell", entry.name)
        self.assertEqual("CLO_0001072", entry.identifier)

        entries = self.manager.get_annotation_entries_by_names(CELL_LINE_URL, ["1321N1 cell"])
        self.assertIsNotNone(entries)
        self.assertEqual(1, len(entries))
        entry = entries[0]
        self.assertEqual("1321N1 cell", entry.name)
        self.assertEqual("CLO_0001072", entry.identifier)

        graph = BELGraph()
        graph.annotation_url[CELL_LINE_KEYWORD] = CELL_LINE_URL

        data = {
            ANNOTATIONS: {
                CELL_LINE_KEYWORD: {
                    "1321N1 cell": True,
                },
            },
        }

        annotations_iter = dict(self.manager._iter_from_annotations_dict(graph, annotations_dict=data[ANNOTATIONS]))
        self.assertIn(CELL_LINE_URL, annotations_iter)
        self.assertIn("1321N1 cell", annotations_iter[CELL_LINE_URL])

        entries = self.manager._get_annotation_entries_from_data(graph, data)
        self.assertIsNotNone(entries)
        self.assertEqual(1, len(entries))
        entry = entries[0]
        self.assertEqual("1321N1 cell", entry.name)
        self.assertEqual("CLO_0001072", entry.identifier)

        self.manager.drop_namespace_by_url(CELL_LINE_URL)
        self.assertEqual(0, self.manager.count_annotations())
        self.assertEqual(0, self.manager.count_annotation_entries())

    def test_get_annotation_entries_no_data(self):
        """Test that if there's no ANNOTATIONS entry in the data, it just returns none."""
        graph = BELGraph()
        data = {}
        entries = self.manager._get_annotation_entries_from_data(graph, data)
        self.assertIsNone(entries)



# -*- coding: utf-8 -*-

"""Tests for Hipathia export."""

import unittest
from typing import Tuple

import pandas as pd

from pybel import BELGraph
from pybel.dsl import ComplexAbundance, Protein
from pybel.io.hipathia import ATT_COLS, to_hipathia_dfs
from pybel.testing.utils import n

PROTEIN_NAMESPACE = "ncbigene"
FAMILY_NAMESPACE = "hipathia.family"
a = Protein(namespace="ncbigene", identifier="P001", name="A")
b_family = Protein(namespace="hipathia.family", identifier="F001", name="B_Family")
b1 = Protein(namespace="ncbigene", identifier="P002", name="B1")
b2 = Protein(namespace="ncbigene", identifier="P003", name="B2")
c_family = Protein(namespace="hipathia.family", identifier="F002", name="C_Family")
c1 = Protein(namespace="ncbigene", identifier="P004", name="C1")
c2 = Protein(namespace="ncbigene", identifier="P005", name="C2")
d = Protein(namespace="ncbigene", identifier="P006", name="D")
c_d = ComplexAbundance([c_family, d])
e = Protein(namespace="ncbigene", identifier="P007", name="E")
f = Protein(namespace="ncbigene", identifier="P008", name="F")
e_f = ComplexAbundance([e, f])

name = "test"


class TestExportHipathia(unittest.TestCase):
    """Test Hipathia."""

    def setUp(self) -> None:
        self.graph = BELGraph(name=name)

    def _get_dfs(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        return to_hipathia_dfs(self.graph)

    def test_protein_activates_protein(self):
        """Test conversion of ``p(A) -> p(D)``."""
        self.graph.add_increases(a, d, citation=n(), evidence="")

        att_df, sif_df = self._get_dfs()

        att_df = att_df[ATT_COLS]
        self.assertEqual(2, len(att_df.index))
        self.assertEqual(
            [
                ["N-{name}-1".format(name=name), a.name, a.identifier],
                ["N-{name}-2".format(name=name), d.name, d.identifier],
            ],
            att_df.values.tolist(),
        )

        self.assertEqual(1, len(sif_df))
        self.assertEqual(
            [
                [
                    "N-{name}-1".format(name=name),
                    "activation",
                    "N-{name}-2".format(name=name),
                ],
            ],
            sif_df.values.tolist(),
        )

    def test_protein_activates_family(self):
        """Test conversion of ``p(A) -> p(B); p(B1) isA p(B); p(B2) isA p(B)``."""
        self.graph.add_increases(a, b_family, citation=n(), evidence=n())
        self.graph.add_is_a(b1, b_family)
        self.graph.add_is_a(b2, b_family)

        att_df, sif_df = self._get_dfs()

        att_df = att_df[ATT_COLS]
        self.assertEqual(2, len(att_df.index))
        self.assertEqual(
            [
                [
                    "N-{name}-1".format(name=name),
                    b_family.name,
                    ",".join((b1.identifier, b2.identifier)),
                ],
                ["N-{name}-2".format(name=name), a.name, a.identifier],
            ],
            att_df.values.tolist(),
        )

        self.assertEqual(1, len(sif_df))
        self.assertEqual(
            [
                [
                    "N-{name}-2".format(name=name),
                    "activation",
                    "N-{name}-1".format(name=name),
                ],
            ],
            sif_df.values.tolist(),
        )

    def test_protein_activates_complex_proteins(self):
        """Test conversion of ``p(A) -> complex(p(E), p(F))``."""
        self.graph.add_increases(a, e_f, citation=n(), evidence="")

        att_df, sif_df = self._get_dfs()

        att_df = att_df[ATT_COLS]
        self.assertEqual(2, len(att_df.index))
        self.assertEqual(
            [
                [
                    "N-{name}-1 2".format(name=name),
                    " ".join((e.name, f.name)),
                    "{},/,{}".format(e.identifier, f.identifier),
                ],
                ["N-{name}-3".format(name=name), a.name, a.identifier],
            ],
            att_df.values.tolist(),
        )

        self.assertEqual(1, len(sif_df))
        self.assertEqual(
            [
                [
                    "N-{name}-3".format(name=name),
                    "activation",
                    "N-{name}-1 2".format(name=name),
                ],
            ],
            sif_df.values.tolist(),
        )

    def test_protein_activates_complex_mixed(self):
        """Test conversion of ``p(A) -> complex(p(E), p(B))`` when b is a family."""
        self.graph.add_increases(a, e_f, citation=n(), evidence="")

        att_df, sif_df = self._get_dfs()

        att_df = att_df[ATT_COLS]
        self.assertEqual(2, len(att_df.index))
        self.assertEqual(
            [
                [
                    "N-{name}-1 2".format(name=name),
                    " ".join((e.name, f.name)),
                    "{},/,{}".format(e.identifier, f.identifier),
                ],
                ["N-{name}-3".format(name=name), a.name, a.identifier],
            ],
            att_df.values.tolist(),
        )

        self.assertEqual(1, len(sif_df))
        self.assertEqual(
            [
                [
                    "N-{name}-3".format(name=name),
                    "activation",
                    "N-{name}-1 2".format(name=name),
                ],
            ],
            sif_df.values.tolist(),
        )



# -*- coding: utf-8 -*-

"""Small graphs with grouped nodes"."""

from ..dsl import Abundance, ComplexAbundance, CompositeAbundance, Protein, Reaction
from ..resources import CHEBI_URL, GO_URL, HGNC_URL
from ..struct.graph import BELGraph

__all__ = [
    "single_reaction_graph",
    "single_composite_graph",
    "single_complex_graph",
]

citation = "None"

evidence = """None""".replace("\n", " ").strip()

single_reaction_graph = BELGraph(
    name="Single Reaction graph",
    version="1.0.0",
    description="Example graph",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

single_reaction_graph.namespace_url.update(
    {
        "HGNC": HGNC_URL,
        "CHEBI": CHEBI_URL,
        "GO": GO_URL,
    }
)

hk1 = Protein(name="HK1", namespace="HGNC", identifier="4922")
atp = Abundance(name="ATP", namespace="CHEBI", identifier="15422")
adp = Abundance(name="ADP", namespace="CHEBI", identifier="16761")
phosphate = Abundance(name="phosphoric acid", namespace="CHEBI", identifier="26078")
glucose = Abundance(name="glucose", namespace="CHEBI", identifier="17234")
glucose_6_phosphate = Abundance(name="D-glucopyranose 6-phosphate", namespace="CHEBI", identifier="4170")
glycolisis_step_1 = Reaction(reactants=[glucose, hk1, atp, phosphate], products=[glucose_6_phosphate, adp, hk1])

composite_example = CompositeAbundance(members=[glucose_6_phosphate, adp, hk1])
complex_example = ComplexAbundance(members=[glucose_6_phosphate, adp, hk1])

single_reaction_graph.add_node_from_data(glycolisis_step_1)

single_complex_graph = BELGraph(
    name="Single Complex graph",
    version="1.0.0",
    description="Example graph",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

single_complex_graph.namespace_url.update(
    {
        "HGNC": HGNC_URL,
        "CHEBI": CHEBI_URL,
        "GO": GO_URL,
    }
)

single_complex_graph.add_node_from_data(complex_example)

single_composite_graph = BELGraph(
    name="Single Composite graph",
    version="1.0.0",
    description="Example graph",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

single_composite_graph.namespace_url.update(
    {
        "HGNC": HGNC_URL,
        "CHEBI": CHEBI_URL,
        "GO": GO_URL,
    }
)

single_composite_graph.add_node_from_data(composite_example)



# -*- coding: utf-8 -*-

"""Exporter for PyNPA.

.. seealso:: https://github.com/pynpa
"""

import logging
import os
from typing import List, Mapping, Optional, Tuple

import pandas as pd

from ..constants import CAUSAL_DECREASE_RELATIONS, CAUSAL_INCREASE_RELATIONS, RELATION
from ..dsl import Gene, MicroRna, Protein, Rna
from ..struct import BELGraph
from ..struct.getters import get_tf_pairs
from ..struct.node_utils import (
    list_abundance_cartesian_expansion,
    reaction_cartesian_expansion,
)

__all__ = [
    "to_npa_directory",
    "to_npa_dfs",
    "to_npa_layers",
]

logger = logging.getLogger(__name__)

Layer = Mapping[Tuple[Gene, Gene], int]

#: Code to distinguish between between iNodes when nodes have been debelized
DEBELIZED_CODE_FOR_INODES = "*"


def to_npa_directory(graph: BELGraph, directory: str, **kwargs) -> None:
    """Write the BEL file to two files in the directory for :mod:`pynpa`."""
    ppi_df, transcription_df = to_npa_dfs(graph, **kwargs)
    ppi_df.to_csv(os.path.join(directory, "ppi_layer.tsv"), sep="\t", index=False)
    transcription_df.to_csv(os.path.join(directory, "transcriptional_layer.tsv"), sep="\t", index=False)


def to_npa_dfs(
    graph: BELGraph,
    cartesian_expansion: bool = False,
    nomenclature_method_first_layer: Optional[str] = None,
    nomenclature_method_second_layer: Optional[str] = None,
    direct_tf_only: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Export the BEL graph as two lists of triples for the :mod:`pynpa`.

    :param graph: A BEL graph
    :param cartesian_expansion: If true, applies cartesian expansion on both reactions (reactants x products)
     as well as list abundances using :func:`list_abundance_cartesian_expansion` and
     :func:`reaction_cartesian_expansion`
    :param nomenclature_method_first_layer: Either "curie", "name" or "inodes. Defaults to "curie".
    :param nomenclature_method_second_layer: Either "curie", "name" or "inodes. Defaults to "curie".

    1. Pick out all transcription factor relationships. Protein X is a transcription
       factor for gene Y IFF ``complex(p(X), g(Y)) -> r(Y)``
    2. Get all other interactions between any gene/rna/protein that are directed causal
       for the PPI layer
    """
    ppi_layer, transcription_layer = to_npa_layers(
        graph,
        cartesian_expansion=cartesian_expansion,
        direct_tf_only=direct_tf_only,
    )
    return (
        _get_df(ppi_layer, method=nomenclature_method_first_layer),
        _get_df(transcription_layer, method=nomenclature_method_second_layer),
    )


def _get_df(layer: Layer, method: Optional[str] = None) -> pd.DataFrame:
    rows = _normalize_layer(layer, method=method)
    return pd.DataFrame(rows, columns=["source", "target", "relation"]).sort_values(["source", "target"])


def _normalize_layer(layer: Layer, method: Optional[str] = None) -> List[Tuple[str, str, int]]:
    if method == "curie" or method is None:
        return [(source.curie, target.curie, direction) for (source, target), direction in layer.items()]
    elif method == "name":
        return [(source.name, target.name, direction) for (source, target), direction in layer.items()]
    elif method == "inodes":
        return [
            (
                "{}{}".format(DEBELIZED_CODE_FOR_INODES, source.name),
                "{}{}".format(DEBELIZED_CODE_FOR_INODES, target.name),
                direction,
            )
            for (source, target), direction in layer.items()
        ]
    else:
        raise ValueError("Invalid export method: {method}".format(method=method))


def to_npa_layers(
    graph: BELGraph,
    cartesian_expansion: bool = False,
    direct_tf_only: bool = False,
) -> Tuple[Layer, Layer]:
    """Get the two layers for the network.

    :param graph: A BEL graph
    :param cartesian_expansion: If true, applies cartesian expansion on both reactions (reactants x products)
     as well as list abundances using :func:`list_abundance_cartesian_expansion` and
     :func:`reaction_cartesian_expansion`
    :param direct_tf_only: If true, only uses directlyIncreases and directlyDecreases relations for TF relations
     ``complex(p(X), g(Y)) =>/=| r(Y)``. If false, also allows indirect relations ``complex(p(X), g(Y)) ->/-| r(Y)``.
    """
    if cartesian_expansion:
        list_abundance_cartesian_expansion(graph)
        reaction_cartesian_expansion(graph)

    transcription_layer = {
        (u.get_rna().get_gene(), v.get_gene()): r for u, v, r in get_tf_pairs(graph, direct_only=direct_tf_only)
    }
    logger.info("extracted %d pairs for the transcription layer", len(transcription_layer))

    ppi_layer = {}
    for u, v, d in graph.edges(data=True):
        u, v = _normalize(u), _normalize(v)
        if u is None or v is None:
            continue
        if (u, v) in transcription_layer:
            continue
        relation = d[RELATION]

        if relation in CAUSAL_INCREASE_RELATIONS:
            ppi_layer[u, v] = +1
        elif relation in CAUSAL_DECREASE_RELATIONS:
            ppi_layer[u, v] = -1
        # TODO what about contradictions

    logger.info("extracted %d pairs for the ppi layer", len(ppi_layer))
    return ppi_layer, transcription_layer


def _normalize(n):
    if isinstance(n, Protein):
        if n.variants:
            n = n.get_parent()
        n = n.get_rna()
    if isinstance(n, (Rna, MicroRna)):
        if n.variants:
            n = n.get_parent()
        n = n.get_gene()
    if isinstance(n, Gene):
        if n.variants:
            n = n.get_parent()
        return n



# -*- coding: utf-8 -*-

"""A module holding the :class:`IdentifierParser`."""

import logging
import re
from collections import defaultdict
from typing import Mapping, Optional, Pattern, Set

from pyparsing import ParseResults, Suppress

from .baseparser import BaseParser
from .constants import NamespaceTermEncodingMapping
from .utils import ns, quote
from ..constants import DIRTY, IDENTIFIER, NAME, NAMESPACE
from ..exceptions import (
    MissingDefaultNameWarning,
    MissingNamespaceNameWarning,
    MissingNamespaceRegexWarning,
    NakedNameWarning,
    UndefinedNamespaceWarning,
)

__all__ = [
    "ConceptParser",
]

logger = logging.getLogger(__name__)


class ConceptParser(BaseParser):
    """A parser for concepts in the form of ``namespace:name`` or ``namespace:identifier!name``.

    Can be made more lenient when given a default namespace or enabling the use of naked names.
    """

    def __init__(
        self,
        namespace_to_term_to_encoding: Optional[NamespaceTermEncodingMapping] = None,
        namespace_to_pattern: Optional[Mapping[str, Pattern]] = None,
        default_namespace: Optional[Set[str]] = None,
        allow_naked_names: bool = False,
        skip_validation: bool = False,
        ensure_go: bool = True,
    ) -> None:
        """Initialize the concept parser.

        :param namespace_to_term_to_encoding: A dictionary of {namespace: {(identifier, name): encoding}}
        :param namespace_to_pattern: A dictionary of {namespace: regular expression string} to compile
        :param default_namespace: A set of strings that can be used without a namespace
        :param allow_naked_names: If true, turn off naked namespace failures
        """
        self.identifier_fqualified = (
            ns(NAMESPACE) + Suppress(":") + (ns | quote)(IDENTIFIER) + Suppress("!") + (ns | quote)(NAME)
        )
        self.identifier_qualified = ns(NAMESPACE) + Suppress(":") + (ns | quote)(NAME)

        if namespace_to_term_to_encoding is not None:
            self.namespace_to_name_to_encoding = defaultdict(dict)
            self.namespace_to_identifier_to_encoding = defaultdict(dict)
            for namespace, term_mapping in namespace_to_term_to_encoding.items():
                for (identifier, name), encoding in term_mapping.items():
                    self.namespace_to_name_to_encoding[namespace][name] = encoding
                    self.namespace_to_identifier_to_encoding[namespace][identifier] = encoding

            self.namespace_to_name_to_encoding = dict(self.namespace_to_name_to_encoding)
            self.namespace_to_identifier_to_encoding = dict(self.namespace_to_identifier_to_encoding)
        else:
            self.namespace_to_name_to_encoding = {}
            self.namespace_to_identifier_to_encoding = {}

        if not skip_validation:
            self.identifier_fqualified.setParseAction(self.handle_identifier_fqualified)
            self.identifier_qualified.setParseAction(self.handle_identifier_qualified)

        self.namespace_to_pattern = namespace_to_pattern or {}
        if ensure_go and "go" not in self.namespace_to_name_to_encoding:
            self.namespace_to_pattern["go"] = re.compile(r"^\d+$")

        self.default_namespace = set(default_namespace) if default_namespace is not None else None
        self.allow_naked_names = allow_naked_names

        self.identifier_bare = (ns | quote)(NAME)
        self.identifier_bare.setParseAction(
            self.handle_namespace_default
            if self.default_namespace
            else self.handle_namespace_lenient
            if self.allow_naked_names
            else self.handle_namespace_invalid,
        )

        super().__init__(
            self.identifier_fqualified | self.identifier_qualified | self.identifier_bare,
        )

    def has_enumerated_namespace(self, namespace: str) -> bool:
        """Check that the namespace has been defined by an enumeration."""
        return namespace in self.namespace_to_name_to_encoding

    def has_regex_namespace(self, namespace: str) -> bool:
        """Check that the namespace has been defined by a regular expression."""
        return namespace in self.namespace_to_pattern

    def raise_for_missing_namespace(self, line: str, position: int, namespace: str, name: str) -> None:
        """Raise an exception if the namespace is not defined."""
        if not self.has_enumerated_namespace(namespace) and not self.has_regex_namespace(namespace):
            raise UndefinedNamespaceWarning(self.get_line_number(), line, position, namespace, name)

    def raise_for_missing_name(self, line: str, position: int, namespace: str, name: str) -> None:
        """Raise an exception if the namespace is not defined or if it does not validate the given name."""
        self.raise_for_missing_namespace(line, position, namespace, name)

        if self.has_enumerated_namespace(namespace) and name not in self.namespace_to_name_to_encoding[namespace]:
            raise MissingNamespaceNameWarning(self.get_line_number(), line, position, namespace, name)

        if self.has_regex_namespace(namespace) and not self.namespace_to_pattern[namespace].match(name):
            raise MissingNamespaceRegexWarning(self.get_line_number(), line, position, namespace, name)

    def handle_identifier_fqualified(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle parsing a qualified OBO-style identifier."""
        return self._handle_identifier(line, position, tokens, key=IDENTIFIER)

    def handle_identifier_qualified(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle parsing a qualified identifier."""
        return self._handle_identifier(line, position, tokens, key=NAME)

    def _handle_identifier(self, line: str, position: int, tokens: ParseResults, key) -> ParseResults:
        """Handle parsing a qualified identifier."""
        namespace, name = tokens[NAMESPACE], tokens[key]

        self.raise_for_missing_namespace(line, position, namespace, name)
        self.raise_for_missing_name(line, position, namespace, name)

        return tokens

    def handle_namespace_default(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle parsing an identifier for the default namespace."""
        name = tokens[NAME]
        if not self.default_namespace:
            raise ValueError("Default namespace is not set")
        if name not in self.default_namespace:
            raise MissingDefaultNameWarning(self.get_line_number(), line, position, name)
        return tokens

    @staticmethod
    def handle_namespace_lenient(line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle parsing an identifier for names missing a namespace that are outside the default namespace."""
        tokens[NAMESPACE] = DIRTY
        logger.debug("Naked namespace: [%d] %s", position, line)
        return tokens

    def handle_namespace_invalid(self, line: str, position: int, tokens: ParseResults) -> None:
        """Raise an exception when parsing a name missing a namespace."""
        name = tokens[NAME]
        raise NakedNameWarning(self.get_line_number(), line, position, name)



# -*- coding: utf-8 -*-

"""Tests for Hipathia."""

import os
import unittest

from pybel.dsl import ComplexAbundance, Protein
from pybel.io.hipathia import from_hipathia_paths, group_delimited_list

HERE = os.path.abspath(os.path.dirname(__file__))

TEST_1_ATT_PATH = os.path.join(HERE, "test_1.att")
TEST_1_SIF_PATH = os.path.join(HERE, "test_1.sif")

TEST_ATT_PATH = os.path.join(HERE, "hsa04370.att")
TEST_SIF_PATH = os.path.join(HERE, "hsa04370.sif")


class TestImportHipathia(unittest.TestCase):
    """Test Hipathia import."""

    def test_import(self):
        """Test importing a hipathia network as a BEL graph."""
        graph = from_hipathia_paths(
            name="test1",
            att_path=TEST_1_ATT_PATH,
            sif_path=TEST_1_SIF_PATH,
        )
        a = Protein(namespace="ncbigene", identifier="1")
        b_family = Protein(namespace="hipathia.family", identifier="B_Family")
        b_2 = Protein(namespace="ncbigene", identifier="2")
        b_3 = Protein(namespace="ncbigene", identifier="3")
        c_family = Protein(namespace="hipathia.family", identifier="C_Family")
        c_4 = Protein(namespace="ncbigene", identifier="4")
        c_5 = Protein(namespace="ncbigene", identifier="5")
        d = Protein(namespace="ncbigene", identifier="6")
        c_d = ComplexAbundance([c_family, d])

        self.assertEqual(
            sorted({a, b_family, b_2, b_3, c_family, c_4, c_5, d, c_d}, key=str),
            sorted(graph, key=str),
        )


class TestUtils(unittest.TestCase):
    def test_group_delimited(self):
        self.assertEqual([[5335, 5336], [9047]], group_delimited_list([5335, 5336, "/", 9047], "/"))



# -*- coding: utf-8 -*-

"""Tests for edge predicates"""

import unittest

from pybel import BELGraph
from pybel.dsl import pathology, protein
from pybel.struct.filters.edge_predicates import has_pathology_causal
from pybel.testing.utils import n


class TestEdgePredicates(unittest.TestCase):
    """Tests for edge predicates."""

    def test_has_pathology(self):
        """Test for checking edges that have a causal pathology."""
        graph = BELGraph()

        a, b, c = protein(n(), n()), pathology(n(), n()), pathology(n(), n())

        key = graph.add_increases(a, b, citation=n(), evidence=n())
        self.assertFalse(has_pathology_causal(graph, a, b, key))

        key = graph.add_increases(b, a, citation=n(), evidence=n())
        self.assertTrue(has_pathology_causal(graph, b, a, key))

        key = graph.add_association(b, a, citation=n(), evidence=n())
        self.assertFalse(has_pathology_causal(graph, b, a, key))

        key = graph.add_increases(a, c, citation=n(), evidence=n())
        self.assertFalse(has_pathology_causal(graph, a, c, key))



# -*- coding: utf-8 -*-

"""This module helps handle node data dictionaries."""

from typing import Any, List, Mapping, Union

from pyparsing import ParseResults

from .constants import (
    CONCEPT,
    FRAGMENT,
    FRAGMENT_DESCRIPTION,
    FRAGMENT_START,
    FRAGMENT_STOP,
    FUNCTION,
    FUSION,
    FUSION_MISSING,
    FUSION_REFERENCE,
    FUSION_START,
    FUSION_STOP,
    GMOD,
    HGVS,
    IDENTIFIER,
    KIND,
    MEMBERS,
    NAME,
    NAMESPACE,
    PARTNER_3P,
    PARTNER_5P,
    PMOD,
    PMOD_CODE,
    PMOD_POSITION,
    PRODUCTS,
    RANGE_3P,
    RANGE_5P,
    REACTANTS,
    REACTION,
    VARIANTS,
    XREFS,
)
from .dsl import (
    FUNC_TO_DSL,
    FUNC_TO_FUSION_DSL,
    FUNC_TO_LIST_DSL,
    BaseAbundance,
    BaseEntity,
    CentralDogma,
    EnumeratedFusionRange,
    Fragment,
    FusionBase,
    FusionRangeBase,
    GeneModification,
    Hgvs,
    ListAbundance,
    MissingFusionRange,
    ProteinModification,
    Reaction,
    Variant,
)

__all__ = [
    "parse_result_to_dsl",
]


def parse_result_to_dsl(tokens) -> BaseEntity:
    """Convert a ParseResult to a PyBEL DSL object.

    :type tokens: dict or pyparsing.ParseResults
    """
    # if MODIFIER in tokens:
    #     return parse_result_to_dsl(tokens[TARGET])
    if REACTION == tokens[FUNCTION]:
        return _reaction_po_to_dict(tokens)

    elif VARIANTS in tokens:
        return _variant_po_to_dict(tokens)

    elif MEMBERS in tokens:
        if CONCEPT in tokens:
            return _list_po_with_concept_to_dict(tokens)
        return _list_po_to_dict(tokens)

    elif FUSION in tokens:
        return _fusion_to_dsl(tokens)

    return _simple_po_to_dict(tokens)


def _fusion_to_dsl(tokens) -> FusionBase:
    """Convert a PyParsing data dictionary to a PyBEL fusion data dictionary.

    :param tokens: A PyParsing data dictionary representing a fusion
    :type tokens: ParseResult
    """
    func = tokens[FUNCTION]
    fusion_dsl = FUNC_TO_FUSION_DSL[func]
    member_dsl = FUNC_TO_DSL[func]

    partner_5p = tokens[FUSION][PARTNER_5P]
    partner_5p_concept = partner_5p[CONCEPT] if CONCEPT in tokens[FUSION][PARTNER_5P] else partner_5p
    partner_5p_node = member_dsl(
        namespace=partner_5p_concept[NAMESPACE],
        name=partner_5p_concept[NAME],
        identifier=partner_5p_concept.get(IDENTIFIER),
        xrefs=partner_5p.get(XREFS),
    )

    partner_3p = tokens[FUSION][PARTNER_3P]
    partner_3p_concept = partner_3p[CONCEPT] if CONCEPT in tokens[FUSION][PARTNER_3P] else partner_3p
    partner_3p_node = member_dsl(
        namespace=partner_3p_concept[NAMESPACE],
        name=partner_3p_concept[NAME],
        identifier=partner_3p_concept.get(IDENTIFIER),
        xrefs=partner_3p.get(XREFS),
    )

    range_5p = _fusion_range_to_dsl(tokens[FUSION][RANGE_5P])
    range_3p = _fusion_range_to_dsl(tokens[FUSION][RANGE_3P])

    return fusion_dsl(
        partner_5p=partner_5p_node,
        partner_3p=partner_3p_node,
        range_5p=range_5p,
        range_3p=range_3p,
    )


def _fusion_range_to_dsl(tokens) -> FusionRangeBase:
    """Convert a PyParsing data dictionary into a PyBEL.

    :type tokens: ParseResult
    """
    if FUSION_MISSING in tokens:
        return MissingFusionRange()

    return EnumeratedFusionRange(
        reference=tokens[FUSION_REFERENCE],
        start=tokens[FUSION_START],
        stop=tokens[FUSION_STOP],
    )


def _simple_po_to_dict(tokens) -> BaseAbundance:
    """Convert a simple named entity to a DSL object.

    :type tokens: ParseResult
    """
    dsl = FUNC_TO_DSL.get(tokens[FUNCTION])
    if dsl is None:
        raise ValueError("invalid tokens: {}".format(tokens))

    concept = tokens[CONCEPT]
    return dsl(
        namespace=concept[NAMESPACE],
        name=concept.get(NAME),
        identifier=concept.get(IDENTIFIER),
        xrefs=tokens.get(XREFS),
    )


def _variant_po_to_dict(tokens) -> CentralDogma:
    """Convert a PyParsing data dictionary to a central dogma abundance (i.e., Protein, RNA, miRNA, Gene).

    :type tokens: ParseResult
    """
    dsl = FUNC_TO_DSL.get(tokens[FUNCTION])
    if dsl is None:
        raise ValueError("invalid tokens: {}".format(tokens))

    concept = tokens[CONCEPT]
    return dsl(
        namespace=concept[NAMESPACE],
        name=concept[NAME],
        identifier=concept.get(IDENTIFIER),
        xrefs=tokens.get(XREFS),
        variants=[_variant_to_dsl_helper(variant_tokens) for variant_tokens in tokens[VARIANTS]],
    )


def _variant_to_dsl_helper(tokens) -> Variant:
    """Convert variant tokens to DSL objects.

    :type tokens: ParseResult
    """
    kind = tokens[KIND]

    if kind == HGVS:
        return Hgvs(tokens[HGVS])

    if kind == GMOD:
        concept = tokens[CONCEPT]
        return GeneModification(
            name=concept[NAME],
            namespace=concept[NAMESPACE],
            identifier=concept.get(IDENTIFIER),
            xrefs=tokens.get(XREFS),
        )

    if kind == PMOD:
        concept = tokens[CONCEPT]
        return ProteinModification(
            name=concept[NAME],
            namespace=concept[NAMESPACE],
            identifier=concept.get(IDENTIFIER),
            xrefs=tokens.get(XREFS),
            code=tokens.get(PMOD_CODE),
            position=tokens.get(PMOD_POSITION),
        )

    if kind == FRAGMENT:
        start, stop = tokens.get(FRAGMENT_START), tokens.get(FRAGMENT_STOP)
        return Fragment(
            start=start,
            stop=stop,
            description=tokens.get(FRAGMENT_DESCRIPTION),
        )

    raise ValueError("invalid fragment kind: {}".format(kind))


def _reaction_po_to_dict(tokens) -> Reaction:
    """Convert a reaction parse object to a DSL.

    :type tokens: ParseResult
    """
    return Reaction(
        reactants=_parse_tokens_list(tokens[REACTANTS]),
        products=_parse_tokens_list(tokens[PRODUCTS]),
    )


def _list_po_with_concept_to_dict(tokens: Union[ParseResults, Mapping[str, Any]]) -> ListAbundance:
    """Convert a list parse object to a node.

    :type tokens: ParseResult
    """
    func = tokens[FUNCTION]
    dsl = FUNC_TO_LIST_DSL[func]
    members = _parse_tokens_list(tokens[MEMBERS])

    concept = tokens[CONCEPT]
    return dsl(
        members=members,
        namespace=concept[NAMESPACE],
        name=concept.get(NAME),
        identifier=concept.get(IDENTIFIER),
        xrefs=tokens.get(XREFS),
    )


def _list_po_to_dict(tokens) -> ListAbundance:
    """Convert a list parse object to a node.

    :type tokens: ParseResult
    """
    func = tokens[FUNCTION]
    dsl = FUNC_TO_LIST_DSL[func]
    members = _parse_tokens_list(tokens[MEMBERS])
    return dsl(members)


def _parse_tokens_list(tokens) -> List[BaseEntity]:
    """Convert a PyParsing result to a reaction.

    :type tokens: ParseResult
    """
    return [parse_result_to_dsl(token) for token in tokens]



# -*- coding: utf-8 -*-

"""Validation for PyBEL data.

The :mod:`pybel.schema` module houses functions to verify the format of a given node or edge.
Its inclusion will help ensure that all PyBEL data is stored in a consistent and
clearly defined manner across the repository.
"""

import json
import logging
import os
import pathlib
from typing import Any, Mapping, Optional, Tuple

import jsonschema

__all__ = ["is_valid_node", "is_valid_edge"]

logger = logging.getLogger(__name__)

HERE = os.path.abspath(os.path.dirname(__file__))

NODE_FILENAME = "base_node.schema.json"
EDGE_FILENAME = "edge.schema.json"

# To use schemas from other files, jsonschema needs to know where the references point to, so
# create a resolver that directs any references (like "entity.schema.json") to the schema's dir
schema_uri = pathlib.PurePath(__file__).as_uri()
RESOLVER = jsonschema.RefResolver(base_uri=schema_uri, referrer=__file__)


def _build_validator(filename: str) -> jsonschema.Draft7Validator:
    """
    Return a validator that checks against a given schema.

    :param filename: The relative path to the schema, e.g. "base_node.schema.json"
    """
    path = os.path.join(HERE, filename)
    with open(path) as json_schema:
        schema = json.load(json_schema)
    return jsonschema.Draft7Validator(schema, resolver=RESOLVER)


def _validate(validator: jsonschema.Draft7Validator, entity: Mapping[str, Any]) -> bool:
    """
    Determine whether a given entity is valid based on its JSON schema.

    :param input: A dict representing a PyBEL entity.
    :return: if the input is valid
    """
    try:
        validator.validate(entity)
        return True
    except jsonschema.exceptions.ValidationError as err:
        logger.info(err)
        return False


node_validator = _build_validator(NODE_FILENAME)
edge_validator = _build_validator(EDGE_FILENAME)


def is_valid_node(node: Mapping[str, Any]) -> bool:
    """
    Determine whether a given node is valid based on the node's JSON schema.

    :param node: A dict representing a PyBEL node.
    :return: if the node is valid
    """
    return _validate(node_validator, node)


def is_valid_edge(edge: Mapping[str, Any]) -> bool:
    """
    Determine whether a given edge is valid based on the edge's JSON schema.

    :param node: A dict representing an edge between two PyBEL nodes.
    :return: if the edge is valid
    """
    return _validate(edge_validator, edge)



# -*- coding: utf-8 -*-

"""Query builder."""

import json
import logging
import random
from collections import UserList
from typing import Any, Dict, List, Set, TextIO, Union

from .constants import (
    SEED_TYPE_ANNOTATION,
    SEED_TYPE_INDUCTION,
    SEED_TYPE_NEIGHBORS,
    SEED_TYPE_SAMPLE,
)
from .selection import get_subgraph
from ...dsl import BaseEntity
from ...struct import union
from ...tokens import parse_result_to_dsl

logger = logging.getLogger(__name__)

SEED_METHOD = "type"
SEED_DATA = "data"

MaybeNodeList = Union[BaseEntity, List[BaseEntity], List[Dict]]


class Seeding(UserList):
    """Represents a container of seeding methods to apply to a network."""

    def append_induction(self, nodes: MaybeNodeList) -> "Seeding":
        """Add a seed induction method.

        :param nodes: A node or list of nodes
        :returns: self for fluid API
        """
        return self._append_seed_handle_nodes(SEED_TYPE_INDUCTION, nodes)

    def append_neighbors(self, nodes: MaybeNodeList) -> "Seeding":
        """Add a seed by neighbors.

        :param nodes: A node or list of nodes
        :returns: self for fluid API
        """
        return self._append_seed_handle_nodes(SEED_TYPE_NEIGHBORS, nodes)

    def append_annotation(self, annotation: str, values: Set[str]) -> "Seeding":
        """Add a seed induction method for single annotation's values.

        :param annotation: The annotation to filter by
        :param values: The values of the annotation to keep
        :returns: self for fluid API
        """
        return self._append_seed(
            SEED_TYPE_ANNOTATION,
            {
                "annotations": {
                    annotation: values,
                },
            },
        )

    def append_sample(self, **kwargs) -> "Seeding":
        """Add seed induction methods.

        Kwargs can have ``number_edges`` or ``number_seed_nodes``.

        :returns: self for fluid API
        """
        data = {
            "seed": random.randint(0, 1000000),
        }
        data.update(kwargs)

        return self._append_seed(SEED_TYPE_SAMPLE, data)

    def _append_seed(self, seed_type: str, data: Any) -> "Seeding":
        """Add a seeding method and returns self.

        :returns: self for fluid API
        """
        self.append(
            {
                SEED_METHOD: seed_type,
                SEED_DATA: data,
            }
        )
        return self

    def _append_seed_handle_nodes(self, seed_type: str, nodes: MaybeNodeList) -> "Seeding":
        """Add a seeding method and returns self.

        :param seed_type: The seed type
        :param nodes: A node or list of nodes
        :returns: self for fluid API
        """
        return self._append_seed(seed_type, _handle_nodes(nodes))

    def run(self, graph):
        """Seed the graph or return none if not possible.

        :type graph: pybel.BELGraph
        :rtype: Optional[pybel.BELGraph]
        """
        if not self:
            logger.debug("no seeding, returning graph: %s", graph)
            return graph

        subgraphs = []

        for seed in self:
            seed_method, seed_data = seed[SEED_METHOD], seed[SEED_DATA]

            logger.debug("seeding with %s: %s", seed_method, seed_data)
            subgraph = get_subgraph(graph, seed_method=seed_method, seed_data=seed_data)

            if subgraph is None:
                logger.debug("seed returned empty graph: %s", seed)
                continue

            subgraphs.append(subgraph)

        if not subgraphs:
            logger.debug("no subgraphs returned")
            return

        return union(subgraphs)

    def to_json(self) -> List[Dict]:
        """Serialize this seeding container to a JSON object."""
        return list(self)

    def dump(self, file, sort_keys: bool = True, **kwargs) -> None:
        """Dump this seeding container to a file as JSON."""
        json.dump(self.to_json(), file, sort_keys=sort_keys, **kwargs)

    def dumps(self, sort_keys: bool = True, **kwargs) -> str:
        """Dump this query to a string as JSON."""
        return json.dumps(self.to_json(), sort_keys=sort_keys, **kwargs)

    @staticmethod
    def from_json(data) -> "Seeding":
        """Build a seeding container from a JSON list."""
        return Seeding(data)

    @staticmethod
    def load(file: TextIO) -> "Seeding":
        """Load a seeding container from a JSON file."""
        return Seeding.from_json(json.load(file))

    @staticmethod
    def loads(s: str) -> "Seeding":
        """Load a seeding container from a JSON string."""
        return Seeding.from_json(json.loads(s))


def _handle_nodes(nodes: MaybeNodeList) -> List[BaseEntity]:
    """Handle node(s) that might be dictionaries."""
    if isinstance(nodes, BaseEntity):
        return [nodes]

    return [(parse_result_to_dsl(node) if not isinstance(node, BaseEntity) else node) for node in nodes]



# -*- coding: utf-8 -*-

"""Input and output functions for BEL graphs.

PyBEL provides multiple lossless interchange options for BEL. Lossy output formats are also included for convenient
export to other programs. Notably, a *de facto* interchange using Resource Description Framework (RDF) to match the
ability of other existing software is excluded due the immaturity of the BEL to RDF mapping.
"""

from .api import dump, load
from .aws import from_s3, to_s3
from .bel_commons_client import from_bel_commons, to_bel_commons
from .biodati_client import from_biodati, to_biodati
from .cx import (
    from_cx,
    from_cx_file,
    from_cx_gz,
    from_cx_jsons,
    to_cx,
    to_cx_file,
    to_cx_gz,
    to_cx_jsons,
)
from .emmaa import from_emmaa
from .extras import to_csv, to_gsea, to_sif
from .fraunhofer_orientdb import from_fraunhofer_orientdb
from .gpickle import (
    from_bytes,
    from_bytes_gz,
    from_pickle,
    from_pickle_gz,
    to_bytes,
    to_bytes_gz,
    to_pickle,
    to_pickle_gz,
)
from .graphdati import (
    from_graphdati,
    from_graphdati_file,
    from_graphdati_gz,
    from_graphdati_jsons,
    to_graphdati,
    to_graphdati_file,
    to_graphdati_gz,
    to_graphdati_jsonl,
    to_graphdati_jsonl_gz,
    to_graphdati_jsons,
)
from .graphml import to_graphml
from .hetionet import (
    from_hetionet_file,
    from_hetionet_gz,
    from_hetionet_json,
    get_hetionet,
)
from .hipathia import (
    from_hipathia_dfs,
    from_hipathia_paths,
    to_hipathia,
    to_hipathia_dfs,
)
from .indra import (
    from_biopax,
    from_indra_pickle,
    from_indra_statements,
    from_indra_statements_json,
    from_indra_statements_json_file,
    to_indra_statements,
    to_indra_statements_json,
    to_indra_statements_json_file,
)
from .jgif import (
    from_cbn_jgif,
    from_cbn_jgif_file,
    from_jgif,
    from_jgif_file,
    from_jgif_gz,
    from_jgif_jsons,
    post_jgif,
    to_jgif,
    to_jgif_file,
    to_jgif_gz,
    to_jgif_jsons,
)
from .jupyter import to_jupyter, to_jupyter_str
from .lines import from_bel_script, from_bel_script_url
from .neo4j import to_neo4j
from .nodelink import (
    from_nodelink,
    from_nodelink_file,
    from_nodelink_gz,
    from_nodelink_jsons,
    to_nodelink,
    to_nodelink_file,
    to_nodelink_gz,
    to_nodelink_jsons,
)
from .pynpa import to_npa_dfs, to_npa_directory
from .sbel import (
    from_sbel,
    from_sbel_file,
    from_sbel_gz,
    to_sbel,
    to_sbel_file,
    to_sbel_gz,
)
from .spia import to_spia_dfs, to_spia_excel, to_spia_tsvs
from .triples import to_edgelist, to_triples, to_triples_file
from .umbrella_nodelink import (
    to_umbrella_nodelink,
    to_umbrella_nodelink_file,
    to_umbrella_nodelink_gz,
)



# -*- coding: utf-8 -*-

"""Filter functions for edges in BEL graphs.

A edge predicate is a function that takes five arguments: a :class:`BELGraph`, a source node, a target node, a key,
and a data dictionary. It returns a boolean representing whether the edge passed the given test.

This module contains a set of default functions for filtering lists of edges and building edge predicate functions.

A general use for an edge predicate is to use the built-in :func:`filter` in code like
:code:`filter(your_edge_predicate, graph.edges(keys=True, data=True))`
"""

from typing import Iterable

from .typing import EdgeIterator, EdgePredicate, EdgePredicates
from ..graph import BELGraph
from ...dsl import BaseEntity

__all__ = [
    "invert_edge_predicate",
    "and_edge_predicates",
    "filter_edges",
    "count_passed_edge_filter",
]


def invert_edge_predicate(edge_predicate: EdgePredicate) -> EdgePredicate:  # noqa: D202
    """Build an edge predicate that is the inverse of the given edge predicate."""

    def _inverse_filter(graph, u, v, k):
        return not edge_predicate(graph, u, v, k)

    return _inverse_filter


def and_edge_predicates(edge_predicates: EdgePredicates) -> EdgePredicate:
    """Concatenate multiple edge predicates to a new predicate that requires all predicates to be met."""
    # If something that isn't a list or tuple is given, assume it's a function and return it
    if not isinstance(edge_predicates, Iterable):
        return edge_predicates

    edge_predicates = tuple(edge_predicates)

    # If only one predicate is given, don't bother wrapping it
    if 1 == len(edge_predicates):
        return edge_predicates[0]

    def concatenated_edge_predicate(graph: BELGraph, u: BaseEntity, v: BaseEntity, k: str) -> bool:
        """Pass only for an edge that pass all enclosed predicates.

        :return: If the edge passes all enclosed predicates
        """
        return all(edge_predicate(graph, u, v, k) for edge_predicate in edge_predicates)

    return concatenated_edge_predicate


def filter_edges(graph: BELGraph, edge_predicates: EdgePredicates) -> EdgeIterator:
    """Apply a set of filters to the edges iterator of a BEL graph.

    :return: An iterable of edges that pass all predicates
    """
    compound_edge_predicate = and_edge_predicates(edge_predicates=edge_predicates)
    for u, v, k in graph.edges(keys=True):
        if compound_edge_predicate(graph, u, v, k):
            yield u, v, k


def count_passed_edge_filter(graph: BELGraph, edge_predicates: EdgePredicates) -> int:
    """Return the number of edges passing a given set of predicates."""
    return sum(1 for _ in filter_edges(graph, edge_predicates=edge_predicates))



# -*- coding: utf-8 -*-

"""The base parser class shared by several BEL parsers."""

import logging
import time
from typing import Iterable, List

from pyparsing import ParserElement, ParseResults

__all__ = ["BaseParser"]

logger = logging.getLogger(__name__)


class BaseParser(object):
    """This abstract class represents a language backed by a PyParsing statement.

    Multiple parsers can be easily chained together when they are all inheriting from this base class.
    """

    def __init__(self, language: ParserElement, streamline: bool = False) -> None:
        """Build a parser wrapper using a PyParsing language.

        :param language: The PyParsing language to use
        :param streamline: Should the language be streamlined on instantiation?
        """
        self.language = language

        #: The parser holds an internal state of the current line
        self._line_number = 0

        if streamline:
            self.streamline()

    def parse_lines(self, lines: Iterable[str]) -> List[ParseResults]:
        """Parse multiple lines in succession."""
        return [self.parseString(line, line_number) for line_number, line in enumerate(lines)]

    def parseString(self, line: str, line_number: int = 0) -> ParseResults:  # noqa: N802
        """Parse a string with the language represented by this parser.

        :param line: A string representing an instance of this parser's language
        :param line_number: The current line number of the parser
        """
        self._line_number = line_number
        return self.language.parseString(line)

    def get_line_number(self) -> int:
        """Get the current line number."""
        return self._line_number

    def streamline(self) -> None:
        """Streamline the language represented by this parser to make queries run faster."""
        t = time.time()
        self.language.streamline()
        logger.info("streamlined %s in %.02f seconds", self.__class__.__name__, time.time() - t)



# -*- coding: utf-8 -*-

"""URLs for default BEL resources.

This script is susceptible to rate limits from the GitHub API, so don't run it over and over!
"""

import logging
import os

from bel_resources.github import get_famplex_url, get_github_url

HERE = os.path.abspath(os.path.dirname(__file__))

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("pybel").setLevel(logging.DEBUG)


def _get_conso_url(name):
    return get_github_url(
        owner="pharmacome",
        repo="conso",
        path="external/{}.belns".format(name),
    )


keyword_to_suffix = dict(
    chebi="chebi-names",
    ec="ec-code",
    fb="fb-names",
    go="go-names",
    hgnc="hgnc-names",
    mesh="mesh-names",
    mgi="mgi-names",
    ncbigene="ncbigene-names",
    rgd="rgd-names",
)


def main():
    """Update the resources links file."""
    keyword_to_url = {keyword: _get_conso_url(suffix) for keyword, suffix in keyword_to_suffix.items()}

    with open(os.path.join(HERE, "resources.py"), "w") as file:
        print("# -*- coding: utf-8 -*-\n", file=file)
        print('"""Resources for PyBEL."""\n', file=file)
        for keyword, url in sorted(keyword_to_url.items()):
            print("{}_URL = '{}'".format(keyword.upper(), url), file=file)

        print("\nFPLX_URL = '{}'".format(get_famplex_url()), file=file)

        print("\n#: Default URL lookup for some keywords", file=file)
        print("keyword_to_url = dict(", file=file)
        for k in sorted(keyword_to_suffix):
            print("    {}={}_URL,".format(k, k.upper()), file=file)
        print("    fplx=FPLX_URL,", file=file)
        print(")", file=file)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Operations for BEL graphs."""

from typing import Iterable

import networkx as nx
from tqdm.autonotebook import tqdm

from .utils import update_metadata
from ..dsl import BaseEntity

__all__ = [
    "subgraph",
    "left_full_join",
    "left_outer_join",
    "union",
    "left_node_intersection_join",
    "node_intersection",
]


def subgraph(graph, nodes: Iterable[BaseEntity]):
    """Induce a sub-graph over the given nodes.

    :rtype: BELGraph
    """
    sg = graph.subgraph(nodes)

    # see implementation for .copy()
    rv = graph.child()
    rv.graph.update(sg.graph)

    for node, data in sg.nodes(data=True):
        rv.add_node(node, **data)

    rv.add_edges_from((u, v, key, datadict.copy()) for u, v, key, datadict in sg.edges(keys=True, data=True))

    return rv


def left_full_join(g, h) -> None:
    """Add all nodes and edges from ``h`` to ``g``, in-place for ``g``.

    :param pybel.BELGraph g: A BEL graph
    :param pybel.BELGraph h: A BEL graph

    Example usage:

    >>> import pybel
    >>> g = pybel.from_bel_script('...')
    >>> h = pybel.from_bel_script('...')
    >>> left_full_join(g, h)
    """
    g.add_nodes_from((node, data) for node, data in h.nodes(data=True) if node not in g)
    g.add_edges_from(
        (u, v, key, data)
        for u, v, key, data in h.edges(keys=True, data=True)
        if u not in g or v not in g[u] or key not in g[u][v]
    )

    update_metadata(h, g)
    g.warnings.extend(h.warnings)


def left_outer_join(g, h) -> None:
    """Only add components from the ``h`` that are touching ``g``.

    Algorithm:

    1. Identify all weakly connected components in ``h``
    2. Add those that have an intersection with the ``g``

    :param BELGraph g: A BEL graph
    :param BELGraph h: A BEL graph

    Example usage:

    >>> import pybel
    >>> g = pybel.from_bel_script('...')
    >>> h = pybel.from_bel_script('...')
    >>> left_outer_join(g, h)
    """
    g_nodes = set(g)
    for comp in nx.weakly_connected_components(h):
        if g_nodes.intersection(comp):
            h_subgraph = subgraph(h, comp)
            left_full_join(g, h_subgraph)


def _left_outer_join_graphs(target, graphs):
    """Outer join a list of graphs to a target graph.

    Note: the order of graphs will have significant results!

    :param BELGraph target: A BEL graph
    :param iter[BELGraph] graphs: An iterator of BEL graphs
    :rtype: BELGraph
    """
    for graph in graphs:
        left_outer_join(target, graph)
    return target


def union(graphs, use_tqdm: bool = False):
    """Take the union over a collection of graphs into a new graph.

    Assumes iterator is longer than 2, but not infinite.

    :param iter[BELGraph] graphs: An iterator over BEL graphs. Can't be infinite.
    :param use_tqdm: Should a progress bar be displayed?
    :return: A merged graph
    :rtype: BELGraph

    Example usage:

    >>> import pybel
    >>> g = pybel.from_bel_script('...')
    >>> h = pybel.from_bel_script('...')
    >>> k = pybel.from_bel_script('...')
    >>> merged = union([g, h, k])
    """
    it = iter(graphs)

    if use_tqdm:
        it = tqdm(it, desc="taking union")

    try:
        target = next(it)
    except StopIteration as e:
        raise ValueError("no graphs given") from e

    try:
        graph = next(it)
    except StopIteration:
        return target
    else:
        target = target.copy()
        left_full_join(target, graph)

    for graph in it:
        left_full_join(target, graph)

    return target


def left_node_intersection_join(g, h):
    """Take the intersection over two graphs.

    This intersection of two graphs is defined by the union of the
    sub-graphs induced over the intersection of their nodes

    :param BELGraph g: A BEL graph
    :param BELGraph h: A BEL graph
    :rtype: BELGraph

    Example usage:

    >>> import pybel
    >>> g = pybel.from_bel_script('...')
    >>> h = pybel.from_bel_script('...')
    >>> merged = left_node_intersection_join(g, h)
    """
    intersecting = set(g).intersection(set(h))

    g_inter = subgraph(g, intersecting)
    h_inter = subgraph(h, intersecting)

    left_full_join(g_inter, h_inter)

    return g_inter


def node_intersection(graphs):
    """Take the node intersection over a collection of graphs into a new graph.

    This intersection is defined the same way as by :func:`left_node_intersection_join`

    :param iter[BELGraph] graphs: An iterable of graphs. Since it's iterated over twice, it gets converted to a
     tuple first, so this isn't a safe operation for infinite lists.
    :rtype: BELGraph

    Example usage:

    >>> import pybel
    >>> g = pybel.from_bel_script('...')
    >>> h = pybel.from_bel_script('...')
    >>> k = pybel.from_bel_script('...')
    >>> merged = node_intersection([g, h, k])
    """
    graphs = tuple(graphs)

    n_graphs = len(graphs)

    if n_graphs == 0:
        raise ValueError("no graphs given")

    if n_graphs == 1:
        return graphs[0]

    nodes = set(graphs[0].nodes())

    for graph in graphs[1:]:
        nodes.intersection_update(graph)

    return union(subgraph(graph, nodes) for graph in graphs)



# -*- coding: utf-8 -*-

"""An example graph in which a family activates another family."""

from ..dsl import Protein
from ..struct import BELGraph

vegf_graph = BELGraph()
vegf = Protein(namespace="fplx", name="VEGF")
vegfr = Protein(namespace="fplx", name="VEGFR")

ev = "VEGF activates the endothelial VEGF receptors" " (VEGFR) 1 and 2, and VEGF-C activates VEGFR-3 and VEGFR-2."

vegf_graph.add_activates(
    vegf,
    vegfr,
    evidence=ev,
    citation="9506953",
)



# -*- coding: utf-8 -*-

"""Utilities for the parsers."""

import itertools as itt
import logging
from typing import Any, List, Optional

from pyparsing import (
    And,
    Group,
    ParserElement,
    Suppress,
    White,
    Word,
    ZeroOrMore,
    alphanums,
    dblQuotedString,
    delimitedList,
    oneOf,
    removeQuotes,
    replaceWith,
)

from ..constants import RELATION, SOURCE, TARGET

logger = logging.getLogger("pybel")


def is_int(s: Any) -> bool:
    """Determine if an object can be cast to an int.

    :param s: any object
    :return: true if argument can be cast to an int:
    """
    try:
        int(s)
        return True
    except ValueError:
        return False


W = Suppress(ZeroOrMore(White()))
C = Suppress(",")
WCW = W + C + W
LPF, RPF = map(Suppress, "()")
LP = Suppress("(") + W
RP = W + Suppress(")")

word = Word(alphanums)
ns = Word(alphanums + "_-.")
identifier = Word(alphanums + "_")
quote = dblQuotedString().setParseAction(removeQuotes)
qid = quote | identifier
delimited_quoted_list = And([Suppress("{"), delimitedList(quote), Suppress("}")])
delimited_unquoted_list = And([Suppress("{"), delimitedList(identifier), Suppress("}")])


def nest(*content):
    """Define a delimited list by enumerating each element of the list."""
    if len(content) == 0:
        raise ValueError("no arguments supplied")
    return And([LPF, content[0]] + list(itt.chain.from_iterable(zip(itt.repeat(C), content[1:]))) + [RPF])


def one_of_tags(
    tags: List[str],
    canonical_tag: str,
    name: Optional[str] = None,
) -> ParserElement:
    """Define the tags usable in the :class:`BelParser`.

    For example, statements like ``g(HGNC:SNCA)`` can be expressed also as ``geneAbundance(HGNC:SNCA)``. The language
    must define multiple different tags that get normalized to the same thing.

    :param tags: a list of strings that are the tags for a function. For example, ['g', 'geneAbundance'] for the
                    abundance of a gene
    :param canonical_tag: the preferred tag name. Does not have to be one of the tags. For example, 'GeneAbundance'
                            (note capitalization) is used for the abundance of a gene
    :param name: this is the key under which the value for this tag is put in the PyParsing framework.
    """
    element = oneOf(tags).setParseAction(replaceWith(canonical_tag))

    if name is None:
        return element

    return element.setResultsName(name)


def triple(subject, relation, obj) -> ParserElement:
    """Build a simple triple in PyParsing that has a ``subject relation object`` format."""
    return And([Group(subject)(SOURCE), relation(RELATION), Group(obj)(TARGET)])



# -*- coding: utf-8 -*-

"""Functions for grouping BEL graphs into sub-graphs."""

from . import annotations, provenance
from .annotations import *
from .provenance import *

__all__ = annotations.__all__ + provenance.__all__



# -*- coding: utf-8 -*-

"""Misc. getters."""

from typing import Iterable, Tuple

from .graph import BELGraph
from ..constants import (
    CAUSAL_DECREASE_RELATIONS,
    CAUSAL_INCREASE_RELATIONS,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    RELATION,
)
from ..dsl import ComplexAbundance, Gene, Protein, Rna

__all__ = [
    "get_tf_pairs",
]


def get_tf_pairs(graph: BELGraph, direct_only: bool = False) -> Iterable[Tuple[Protein, Rna, int]]:
    """Iterate pairs of ``p(X)`` and ``r(Y)`` such that ``complex(p(X), g(Y)) -> r(Y)``.

    :param graph: A BEL graph
    :param direct_only: If true, only uses directlyIncreases and directlyDecreases relations. Otherwise, allows
     indirect relations.
    """
    if direct_only:
        _inc, _dec = {DIRECTLY_INCREASES}, {DIRECTLY_DECREASES}
    else:
        _inc, _dec = CAUSAL_INCREASE_RELATIONS, CAUSAL_DECREASE_RELATIONS

    for tf in _iterate_proteins(graph):
        for tf_gene in graph[tf]:
            if not isinstance(tf_gene, ComplexAbundance):
                continue
            if tf not in tf_gene.members:
                continue
            other_members = [m for m in tf_gene.members if m != tf]
            if 1 != len(other_members):
                continue
            target_gene = other_members[0]
            if not isinstance(target_gene, Gene):
                continue
            if target_gene.variants:
                target_gene = target_gene.get_parent()
            target_rna = target_gene.get_rna()
            if target_rna not in graph:
                continue
            for edge in graph[tf_gene][target_rna].values():
                relation = edge[RELATION]
                if relation in _inc:
                    yield tf, target_rna, +1
                elif relation in _dec:
                    yield tf, target_rna, -1


def _iterate_proteins(graph: BELGraph) -> Iterable[Protein]:
    return (node for node in graph if isinstance(node, Protein))



# -*- coding: utf-8 -*-

"""Tests for collapse functions."""

import unittest

from pybel import BELGraph
from pybel.constants import DIRECTLY_INCREASES
from pybel.dsl import gene, mirna, pathology, pmod, protein, rna
from pybel.struct.mutation.collapse import (
    collapse_all_variants,
    collapse_nodes,
    collapse_to_genes,
    surviors_are_inconsistent,
)
from pybel.testing.utils import n

HGNC = "HGNC"
GO = "GO"
CHEBI = "CHEBI"

g1 = gene(HGNC, "1")
r1 = rna(HGNC, "1")
p1 = protein(HGNC, "1")
p1_phosphorylated = protein(HGNC, "1", variants=[pmod("Ph")])

g2 = gene(HGNC, "2")
r2 = rna(HGNC, "2")
p2 = protein(HGNC, "2")

g3 = gene(HGNC, "3")
r3 = rna(HGNC, "3")
p3 = protein(HGNC, "3")

g4 = gene(HGNC, "4")
m4 = mirna(HGNC, "4")

p5 = pathology(GO, "5")


class TestCollapse(unittest.TestCase):
    """Tests for collapse functions."""

    def test_check_survivors_consistent(self):
        """Test the survivor mapping is consistent."""
        inconsistencies = surviors_are_inconsistent(
            {
                1: {2},
                3: {4},
            }
        )
        self.assertEqual(0, len(inconsistencies))
        self.assertFalse(inconsistencies)

        inconsistencies = surviors_are_inconsistent(
            {
                1: {2},
                2: {3},
                3: {4},
                5: {4},
            }
        )
        self.assertEqual(2, len(inconsistencies))
        self.assertIn(2, inconsistencies)
        self.assertIn(3, inconsistencies)

    def test_collapse_by_dict(self):
        """Test collapsing nodes by a dictionary."""
        graph = BELGraph()
        graph.add_node_from_data(p1)
        graph.add_node_from_data(p2)
        graph.add_node_from_data(p3)

        graph.add_increases(p1, p3, citation=n(), evidence=n())
        graph.add_qualified_edge(p2, p3, relation=DIRECTLY_INCREASES, citation=n(), evidence=n())

        self.assertEqual(3, graph.number_of_nodes())
        self.assertEqual(2, graph.number_of_edges())

        d = {p1: {p2}}

        collapse_nodes(graph, d)

        self.assertEqual({p1, p3}, set(graph))
        self.assertEqual({(p1, p3), (p1, p3)}, set(graph.edges()))
        self.assertEqual(2, graph.number_of_edges(), msg=graph.edges(data=True, keys=True))

    def test_collapse_dogma_1(self):
        """Test collapsing to genes, only with translations."""
        graph = BELGraph()
        graph.add_translation(r1, p1)

        self.assertEqual(2, graph.number_of_nodes())
        self.assertEqual(1, graph.number_of_edges())

        collapse_to_genes(graph)

        self.assertIn(g1, graph)
        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())

    def test_collapse_dogma_2(self):
        """Test collapsing to genes with translations and transcriptions."""
        graph = BELGraph()
        graph.add_transcription(g1, r1)
        graph.add_translation(r1, p1)

        self.assertEqual(3, graph.number_of_nodes())
        self.assertEqual(2, graph.number_of_edges())

        collapse_to_genes(graph)

        self.assertIn(g1, graph)
        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())

    def test_collapse_dogma_3(self):
        """Test collapsing to genes, only with transcriptions."""
        graph = BELGraph()
        graph.add_transcription(g1, r1)

        self.assertEqual(2, graph.number_of_nodes())
        self.assertEqual(1, graph.number_of_edges())

        collapse_to_genes(graph)

        self.assertIn(g1, graph)
        self.assertEqual(1, graph.number_of_nodes())
        self.assertEqual(0, graph.number_of_edges())

    def test_collapse_all_variants(self):
        """Test collapsing all variants to their reference nodes."""
        graph = BELGraph()
        graph.add_node_from_data(p1_phosphorylated)

        graph.add_increases(p1_phosphorylated, p2, citation=n(), evidence=n())

        self.assertEqual(3, graph.number_of_nodes())
        self.assertEqual(2, graph.number_of_edges())

        collapse_all_variants(graph)

        self.assertEqual(2, graph.number_of_nodes())
        self.assertEqual(1, graph.number_of_edges())

        self.assertIn(p1, graph)
        self.assertNotIn(p1_phosphorylated, graph)
        self.assertIn(p2, graph)



# -*- coding: utf-8 -*-

"""Constants for the query builder."""

#: Induce a subgraph over the given nodes
SEED_TYPE_INDUCTION = "induction"
#: Induce a subgraph over the given nodes and expand to their first neighbors
SEED_TYPE_NEIGHBORS = "neighbors"
#: Induce a subgraph over the given nodes and expand to their second neighbors
SEED_TYPE_DOUBLE_NEIGHBORS = "dneighbors"
#: Induce a subgraph over the nodes in all shortest paths between the given nodes
SEED_TYPE_PATHS = "shortest_paths"
#: Induce a subgraph over the edges provided by the given authors and their neighboring nodes
SEED_TYPE_AUTHOR = "authors"
#: Induce a subgraph over the edges provided by the given citations and their neighboring nodes
SEED_TYPE_PUBMED = "pubmed"
#: Generate an upstream candidate mechanism
SEED_TYPE_UPSTREAM = "upstream"
#: Generate a downstream candidate mechanism
SEED_TYPE_DOWNSTREAM = "downstream"
#: Induce a subgraph over the edges matching the given annotations
SEED_TYPE_ANNOTATION = "annotation"
#: Induce a subgraph over a random set of (hopefully) connected edges
SEED_TYPE_SAMPLE = "sample"

#: A set of the allowed seed type strings, as defined above
SEED_TYPES = {
    SEED_TYPE_INDUCTION,
    SEED_TYPE_NEIGHBORS,
    SEED_TYPE_DOUBLE_NEIGHBORS,
    SEED_TYPE_PATHS,
    SEED_TYPE_UPSTREAM,
    SEED_TYPE_DOWNSTREAM,
    SEED_TYPE_PUBMED,
    SEED_TYPE_AUTHOR,
    SEED_TYPE_ANNOTATION,
    SEED_TYPE_SAMPLE,
}

#: Seed types that don't take node lists as their arguments
NONNODE_SEED_TYPES = {
    SEED_TYPE_ANNOTATION,
    SEED_TYPE_AUTHOR,
    SEED_TYPE_PUBMED,
    SEED_TYPE_SAMPLE,
}

NODE_SEED_TYPES = SEED_TYPES - NONNODE_SEED_TYPES



# -*- coding: utf-8 -*-

"""A manager for looking up nodes."""

from typing import List, Optional

from sqlalchemy import and_

from .base_manager import BaseManager
from .models import Author, Citation, Edge, Evidence, Node
from ..constants import CITATION_TYPE_PUBMED
from ..dsl import BaseEntity


class LookupManager(BaseManager):
    """Groups functions for looking up entries by hashes."""

    def get_dsl_by_hash(self, node_hash: str) -> Optional[BaseEntity]:
        """Look up a node by the hash and returns the corresponding PyBEL node tuple."""
        node = self.get_node_by_hash(node_hash)
        if node is not None:
            return node.as_bel()

    def get_node_by_hash(self, node_hash: str) -> Optional[Node]:
        """Look up a node by its hash."""
        return self.session.query(Node).filter(Node.md5 == node_hash).one_or_none()

    def get_nodes_by_hashes(self, node_hashes: List[str]) -> List[Node]:
        """Look up several nodes by their hashes."""
        return self.session.query(Node).filter(Node.md5.in_(node_hashes)).all()

    def get_node_by_dsl(self, node: BaseEntity) -> Optional[Node]:
        """Look up a node by its data dictionary by hashing it then using :func:`get_node_by_hash`."""
        return self.get_node_by_hash(node.md5)

    def get_edge_by_hash(self, edge_hash: str) -> Optional[Edge]:
        """Look up an edge by the hash of a PyBEL edge data dictionary."""
        return self.session.query(Edge).filter(Edge.md5 == edge_hash).one_or_none()

    def get_edges_by_hashes(self, edge_hashes: List[str]) -> List[Edge]:
        """Look up several edges by hashes of their PyBEL edge data dictionaries."""
        return self.session.query(Edge).filter(Edge.md5.in_(edge_hashes)).all()

    def get_citation_by_pmid(self, pubmed_identifier: str) -> Optional[Citation]:
        """Get a citation object by its PubMed identifier."""
        return self.get_citation_by_reference(db_id=pubmed_identifier, db=CITATION_TYPE_PUBMED)

    def get_citation_by_reference(self, db: str, db_id: str) -> Optional[Citation]:
        """Get a citation object by its database and reference."""
        return self.session.query(Citation).filter(Citation.db == db, Citation.db_id == db_id).one_or_none()

    def get_citation_by_curie(self, curie: str) -> Optional[Citation]:
        """Get a citation object by its hash."""
        db, db_id = curie.split(":")
        return self.get_citation_by_reference(db=db, db_id=db_id)

    def get_author_by_name(self, name: str) -> Optional[Author]:
        """Get an author by name, if it exists in the database."""
        return self.session.query(Author).filter(Author.name == name).one_or_none()

    def get_evidence_by_hash(self, evidence_hash: str) -> Optional[Evidence]:
        """Look up an evidence by its hash."""
        return self.session.query(Evidence).filter(Evidence.md5 == evidence_hash).one_or_none()

    def get_evidence_by_reference_text(self, db: str, db_id: str, text: str) -> Optional[Evidence]:
        """Look up an evidence by its citation's database/identifier and text."""
        citation = self.get_citation_by_reference(db=db, db_id=db_id)
        if citation is not None:
            return self.get_evidence_by_citation_text(citation, text)

    def get_evidence_by_citation_text(self, citation: Citation, text: str) -> Optional[Evidence]:
        """Look up an evidence by its citation and text."""
        f = and_(Evidence.citation == citation, Evidence.text == text)
        return self.session.query(Evidence).filter(f).one_or_none()



# -*- coding: utf-8 -*-

"""Utilities for displaying graphs with inline HTML in Jupyter Notebooks."""

from random import sample
from typing import Mapping, Optional

from .constants import DEFAULT_COLOR_MAP
from ..jinja_utils import build_template_renderer
from ..nodelink import to_nodelink_jsons
from ...struct import BELGraph

__all__ = ["to_jupyter", "to_jupyter_str"]

DEFAULT_WIDTH = 1000
DEFAULT_HEIGHT = 650


def _generate_id() -> str:
    """Generate a random string of letters."""
    return "".join(sample("abcdefghjkmopqrstuvqxyz", 16))


def to_jupyter(
    graph: BELGraph,
    width: int = DEFAULT_WIDTH,
    height: int = DEFAULT_HEIGHT,
    color_map: Optional[Mapping[str, str]] = None,
):
    """Display a BEL graph inline in a Jupyter notebook.

    To use successfully, make run as the last statement in a cell inside a Jupyter notebook.

    :param graph: A BEL graph
    :param width: The width of the visualization window to render
    :param height: The height of the visualization window to render
    :param color_map: A dictionary from PyBEL internal node functions to CSS color strings like #FFEE00. Defaults
                    to :data:`default_color_map`
    :return: An IPython notebook Javascript object
    :rtype: :class:`IPython.display.Javascript`
    """
    from IPython.display import Javascript

    return Javascript(
        to_jupyter_str(
            graph,
            width=width,
            height=height,
            color_map=color_map,
        )
    )


def to_jupyter_str(
    graph: BELGraph,
    width: int = DEFAULT_WIDTH,
    height: int = DEFAULT_HEIGHT,
    color_map: Optional[Mapping[str, str]] = None,
) -> str:
    """Return the string to be javascript-ified by the Jupyter notebook function :class:`IPython.display.Javascript`.

    :param graph: A BEL graph
    :param width: The width of the visualization window to render
    :param height: The height of the visualization window to render
    :param color_map: A dictionary from PyBEL internal node functions to CSS color strings like #FFEE00. Defaults
                    to :data:`default_color_map`
    :return: The javascript string to turn into magic
    """
    gjson = to_nodelink_jsons(graph)
    chart_id = _generate_id()

    #: Renders templates from pybel.io.jupyter.templates folder
    render_template = build_template_renderer(__file__)

    return render_template(
        "pybel_jupyter.js",
        graph=gjson,
        chart=chart_id,
        width=width,
        height=height,
        color_map=(color_map or DEFAULT_COLOR_MAP),
    )



# -*- coding: utf-8 -*-

"""An example graph in which a famplex (complex of families) activates something."""

from ..dsl import ComplexAbundance, NamedComplexAbundance, Protein, ProteinModification
from ..struct import BELGraph

ampk_graph = BELGraph()
ampk = NamedComplexAbundance(namespace="fplx", name="AMPK")

# Alpha subunits of AMPK
ampk_alpha = Protein(namespace="fplx", name="AMPK_alpha")
ampk_graph.add_part_of(ampk_alpha, ampk)

prkaa1 = Protein(namespace="hgnc", identifier="9376", name="PRKAA1")
prkaa2 = Protein(namespace="hgnc", identifier="9377", name="PRKAA2")
ampk_graph.add_is_a(prkaa1, ampk_alpha)
ampk_graph.add_is_a(prkaa2, ampk_alpha)

# Beta subunits of AMPK
ampk_beta = Protein(namespace="fplx", name="AMPK_beta")
ampk_graph.add_part_of(ampk_beta, ampk)

prkab1 = Protein(namespace="hgnc", identifier="9378", name="PRKAB1")
prkab2 = Protein(namespace="hgnc", identifier="9379", name="PRKAB2")
ampk_graph.add_is_a(prkab1, ampk_beta)
ampk_graph.add_is_a(prkab2, ampk_beta)

# Gamma subunits of AMPK
ampk_gamma = Protein(namespace="fplx", name="AMPK_gamma")
ampk_graph.add_part_of(ampk_gamma, ampk)

prkag1 = Protein(namespace="hgnc", identifier="9385", name="PRKAG1")
prkag2 = Protein(namespace="hgnc", identifier="9386", name="PRKAG2")
prkag3 = Protein(namespace="hgnc", identifier="9387", name="PRKAG3")
ampk_graph.add_is_a(prkag1, ampk_gamma)
ampk_graph.add_is_a(prkag2, ampk_gamma)
ampk_graph.add_is_a(prkag3, ampk_gamma)

mtorc1 = NamedComplexAbundance(namespace="fplx", name="mTORC1")
mtor = Protein(namespace="hgnc", identifier="3942", name="MTOR")
rptor = Protein(namespace="hgnc", identifier="30287", name="RPTOR")
ampk_graph.add_part_of(mtor, mtorc1)
ampk_graph.add_part_of(rptor, mtorc1)

# FamPlex says this is a family but I thought it was a complex
p14_3_3 = Protein(namespace="fplx", name="p14_3_3")

ev = (
    "We report here that AMPK directly phosphorylates the"
    " mTOR binding partner raptor on two well conserved serine"
    " residues, and this phosphorylation induces 14-3-3"
    " binding to raptor."
)

ampk_graph.add_directly_phosphorylates(
    ampk,
    rptor,
    "Ser",  # on Ser722 and Ser792
    evidence=ev,
    citation="18439900",
)
ampk_graph.add_binds(
    ComplexAbundance(
        [
            rptor.with_variants(ProteinModification("Ph")),
            mtor,
        ]
    ),
    p14_3_3,
    evidence=ev,
    citation="18439900",
)



# -*- coding: utf-8 -*-

"""Functions for expanding the neighborhoods of nodes."""

import itertools as itt
from typing import Iterable

from ...filters.node_predicates import is_pathology
from ...filters.typing import EdgeIterator
from ...graph import BELGraph
from ...pipeline import uni_in_place_transformation
from ...utils import update_metadata
from ....dsl import BaseEntity

__all__ = [
    "expand_node_predecessors",
    "expand_node_successors",
    "expand_node_neighborhood",
    "expand_nodes_neighborhoods",
    "expand_all_node_neighborhoods",
    "expand_internal",
]


@uni_in_place_transformation
def expand_node_predecessors(universe: BELGraph, graph: BELGraph, node: BaseEntity):
    """Expand around the predecessors of the given node in the result graph.

    :param universe: The graph containing the stuff to add
    :param graph: The graph to add stuff to
    :param node: A BEL node
    """
    skip_successors = set()
    for successor in universe.successors(node):
        if successor in graph:
            skip_successors.add(successor)
            continue

        graph.add_node_from_data(successor)

    graph.add_edges_from(
        (source, successor, key, data)
        for source, successor, key, data in universe.out_edges(node, data=True, keys=True)
        if successor not in skip_successors
    )
    update_metadata(universe, graph)


@uni_in_place_transformation
def expand_node_successors(universe: BELGraph, graph: BELGraph, node: BaseEntity) -> None:
    """Expand around the successors of the given node in the result graph.

    :param universe: The graph containing the stuff to add
    :param graph: The graph to add stuff to
    :param node: A BEL node
    """
    skip_predecessors = set()
    for predecessor in universe.predecessors(node):
        if predecessor in graph:
            skip_predecessors.add(predecessor)
            continue

        graph.add_node_from_data(predecessor)

    graph.add_edges_from(
        (predecessor, target, key, data)
        for predecessor, target, key, data in universe.in_edges(node, data=True, keys=True)
        if predecessor not in skip_predecessors
    )

    update_metadata(universe, graph)


@uni_in_place_transformation
def expand_node_neighborhood(universe: BELGraph, graph: BELGraph, node: BaseEntity) -> None:
    """Expand around the neighborhoods of the given node in the result graph.

    Note: expands complexes' members

    :param universe: The graph containing the stuff to add
    :param graph: The graph to add stuff to
    :param node: A BEL node
    """
    expand_node_predecessors(universe, graph, node)
    expand_node_successors(universe, graph, node)


@uni_in_place_transformation
def expand_nodes_neighborhoods(universe: BELGraph, graph: BELGraph, nodes: Iterable[BaseEntity]) -> None:
    """Expand around the neighborhoods of the given node in the result graph.

    :param universe: The graph containing the stuff to add
    :param graph: The graph to add stuff to
    :param nodes: Nodes from the query graph
    """
    for node in nodes:
        expand_node_neighborhood(universe, graph, node)


@uni_in_place_transformation
def expand_all_node_neighborhoods(universe: BELGraph, graph: BELGraph, filter_pathologies: bool = False) -> None:
    """Expand the neighborhoods of all nodes in the given graph.

    :param pybel.BELGraph universe: The graph containing the stuff to add
    :param pybel.BELGraph  graph: The graph to add stuff to
    :param filter_pathologies: Should expansion take place around pathologies?
    """
    for node in list(graph):
        if filter_pathologies and is_pathology(node):
            continue

        expand_node_neighborhood(universe, graph, node)


@uni_in_place_transformation
def expand_internal(
    universe: BELGraph,
    graph: BELGraph,
) -> None:
    """Expand on edges between entities in the sub-graph that pass the given filters, in place.

    :param universe: The full graph
    :param graph: A sub-graph to find the upstream information
    """
    for u, v, key in iterate_internal(universe, graph):
        graph.add_edge(u, v, key=key, **universe[u][v][key])


def iterate_internal(universe: BELGraph, graph: BELGraph) -> EdgeIterator:
    """Iterate over edges that are in the universe but not the target graph."""
    for u, v in itt.product(graph, repeat=2):
        if graph.has_edge(u, v) or not universe.has_edge(u, v):
            continue
        for key in universe[u][v]:
            yield u, v, key



# -*- coding: utf-8 -*-

"""This module tests the to_json functions for all of the database models."""

import datetime
import json
import unittest

from pybel.constants import (
    CITATION_TYPE_PUBMED,
    IDENTIFIER,
    METADATA_AUTHORS,
    METADATA_CONTACT,
    METADATA_COPYRIGHT,
    METADATA_DESCRIPTION,
    METADATA_DISCLAIMER,
    METADATA_LICENSES,
    METADATA_NAME,
    METADATA_VERSION,
    NAME,
    NAMESPACE,
    NAMESPACE_DOMAIN_OTHER,
)
from pybel.language import citation_dict
from pybel.manager.models import Citation, Namespace, NamespaceEntry, Network
from pybel.testing.utils import n


class TestNetwork(unittest.TestCase):
    def setUp(self):
        self.name = n()
        self.version = n()
        self.created = datetime.datetime.utcnow()

        self.model = Network(
            name=self.name,
            version=self.version,
            created=self.created,
        )
        self.expected = {
            METADATA_NAME: self.name,
            METADATA_VERSION: self.version,
            "created": str(self.created),
        }

    def test_to_json(self):
        model_json = self.model.to_json()

        self.assertIn(METADATA_NAME, model_json)
        self.assertEqual(self.name, model_json[METADATA_NAME])
        self.assertIn(METADATA_VERSION, model_json)
        self.assertEqual(self.version, model_json[METADATA_VERSION])
        self.assertIn("created", model_json)
        self.assertEqual(str(self.created), model_json["created"])

        self.assertEqual(self.expected, model_json)

    def test_dump(self):
        json.dumps(self.model)

    def test_network(self):
        self.expected[METADATA_AUTHORS] = self.model.authors = n()
        self.assertEqual(self.expected, self.model.to_json())

        self.expected[METADATA_CONTACT] = self.model.contact = n()
        self.assertEqual(self.expected, self.model.to_json())

        self.expected[METADATA_DESCRIPTION] = self.model.description = n()
        self.assertEqual(self.expected, self.model.to_json())

        self.expected[METADATA_COPYRIGHT] = self.model.copyright = n()
        self.assertEqual(self.expected, self.model.to_json())

        self.expected[METADATA_DISCLAIMER] = self.model.disclaimer = n()
        self.assertEqual(self.expected, self.model.to_json())

        self.expected[METADATA_LICENSES] = self.model.licenses = n()
        self.assertEqual(self.expected, self.model.to_json())

        self.expected["id"] = None
        self.assertEqual(self.expected, self.model.to_json(include_id=True))


class TestModels(unittest.TestCase):
    def test_namespace_url(self):
        uploaded = datetime.datetime.now()

        model = Namespace(
            keyword="TEST",
            url="http://test.com",
            name="Test Namespace",
            domain=NAMESPACE_DOMAIN_OTHER,
            species="9606",
            version="1.0.0",
            author="Charles Tapley Hoyt",
            contact="cthoyt@gmail.com",
            uploaded=uploaded,
        )

        expected = dict(
            keyword="TEST",
            url="http://test.com",
            name="Test Namespace",
            version="1.0.0",
        )

        self.assertEqual(model.to_json(), expected)

        expected["id"] = model.id = 1

        self.assertEqual(model.to_json(include_id=True), expected)

    def test_namespace_pattern(self):
        uploaded = datetime.datetime.now()

        model = Namespace(
            keyword="TEST",
            pattern=r"\w+",
            name="Test Namespace",
            domain=NAMESPACE_DOMAIN_OTHER,
            species="9606",
            version="1.0.0",
            author="Charles Tapley Hoyt",
            contact="cthoyt@gmail.com",
            uploaded=uploaded,
        )

        expected = dict(
            keyword="TEST",
            pattern=r"\w+",
            name="Test Namespace",
            version="1.0.0",
        )

        self.assertEqual(model.to_json(), expected)

    def test_namespace_entry(self):
        model = NamespaceEntry(name="entry", namespace=Namespace(keyword="test"))

        expected = {
            NAMESPACE: "test",
            NAME: "entry",
        }

        self.assertEqual(model.to_json(), expected)

        expected["id"] = model.id = 1

        self.assertEqual(model.to_json(include_id=True), expected)

        expected[IDENTIFIER] = model.identifier = "test:00001"

        self.assertEqual(model.to_json(include_id=True), expected)

    def test_citation(self):
        db_id = n()
        model = Citation(
            db=CITATION_TYPE_PUBMED,
            db_id=db_id,
        )

        expected = citation_dict(namespace=CITATION_TYPE_PUBMED, identifier=db_id)
        self.assertEqual(expected, model.to_json())

        expected[NAME] = model.title = n()
        self.assertEqual(expected, model.to_json())



# -*- coding: utf-8 -*-

"""Curation of the article "Genetics ignite focus on microglial inflammation in Alzheimer's disease".

.. code-block:: none

    SET Citation = {"PubMed", "26438529"}
    SET Evidence = "Sialic acid binding activates CD33, resulting in phosphorylation of the CD33
    immunoreceptor tyrosine-based inhibitory motif (ITIM) domains and activation of the SHP-1 and
    SHP-2 tyrosine phosphatases [66, 67]."

    complex(p(HGNC:CD33),a(CHEBI:"sialic acid")) -> p(HGNC:CD33, pmod(P))
    act(p(HGNC:CD33, pmod(P))) => act(p(HGNC:PTPN6), ma(phos))
    act(p(HGNC:CD33, pmod(P))) => act(p(HGNC:PTPN11), ma(phos))

    UNSET {Evidence, Species}

    SET Evidence = "These phosphatases act on multiple substrates, including Syk, to inhibit immune
    activation [68, 69].  Hence, CD33 activation leads to increased SHP-1 and SHP-2 activity that antagonizes Syk,
    inhibiting ITAM-signaling proteins, possibly including TREM2/DAP12 (Fig. 1, [70, 71])."

    SET Species = 9606

    act(p(HGNC:PTPN6)) =| act(p(HGNC:SYK))
    act(p(HGNC:PTPN11)) =| act(p(HGNC:SYK))
    act(p(HGNC:SYK)) -> act(p(HGNC:TREM2))
    act(p(HGNC:SYK)) -> act(p(HGNC:TYROBP))

    UNSET ALL
"""

from ..dsl import (
    Abundance,
    BiologicalProcess,
    ComplexAbundance,
    Entity,
    Protein,
    ProteinModification,
    activity,
)
from ..resources import CHEBI_URL, CONFIDENCE_URL, GO_URL, HGNC_URL, SPECIES_PATTERN
from ..struct.graph import BELGraph

__all__ = [
    "sialic_acid_graph",
]

citation = "26438529"

evidence_1 = """
Sialic acid binding activates CD33, resulting in phosphorylation of the CD33 immunoreceptor tyrosine-based inhibitory
motif (ITIM) domains and activation of the SHP-1 and SHP-2 tyrosine phosphatases [66, 67].
""".replace(
    "\n", " "
).strip()

evidence_2 = """These phosphatases act on multiple substrates, including Syk, to inhibit immune activation [68, 69].
Hence, CD33 activation leads to increased SHP-1 and SHP-2 activity that antagonizes Syk, inhibiting ITAM-signaling
proteins, possibly including TREM2/DAP12 (Fig. 1, [70, 71]).
""".replace(
    "\n", " "
).strip()

sialic_acid_graph = BELGraph(
    name="Sialic Acid Graph",
    version="1.0.0",
    description="The downstream effects of sialic acid in immune signaling",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

sialic_acid_graph.namespace_url.update(
    {
        "hgnc": HGNC_URL,
        "chebi": CHEBI_URL,
        "go": GO_URL,
    }
)

sialic_acid_graph.annotation_url.update(
    {
        "Confidence": CONFIDENCE_URL,
    }
)

sialic_acid_graph.annotation_pattern.update(
    {
        "Species": SPECIES_PATTERN,
    }
)

sialic_acid = Abundance(name="sialic acid", namespace="chebi", identifier="26667")
cd33 = Protein(
    name="CD33",
    namespace="hgnc",
    identifier="1659",
    xrefs=[
        Entity(namespace="uniprot", identifier="P20138"),
    ],
)
sialic_acid_cd33_complex = ComplexAbundance([sialic_acid, cd33])
shp1 = Protein(namespace="hgnc", name="PTPN6", identifier="9658")
shp2 = Protein(namespace="hgnc", name="PTPN11", identifier="9644")
syk = Protein(namespace="hgnc", name="SYK", identifier="11491")
dap12 = Protein(namespace="hgnc", name="TYROBP", identifier="12449")
trem2 = Protein(namespace="hgnc", name="TREM2", identifier="17761")
cd33_phosphorylated = Protein(name="CD33", namespace="hgnc", identifier="1659", variants=ProteinModification("Ph"))
immune_response = BiologicalProcess(name="immune response", namespace="go", identifier="0006955")

sialic_acid_graph.add_increases(
    sialic_acid_cd33_complex,
    cd33,
    citation=citation,
    annotations={"Species": "9606", "Confidence": "High"},
    evidence=evidence_1,
    target_modifier=activity(),
)

sialic_acid_graph.add_increases(
    cd33,
    cd33_phosphorylated,
    citation=citation,
    annotations={"Species": "9606", "Confidence": "High"},
    evidence=evidence_1,
    source_modifier=activity(),
)

sialic_acid_graph.add_directly_increases(
    cd33_phosphorylated,
    shp1,
    citation=citation,
    evidence=evidence_1,
    annotations={"Species": "9606", "Confidence": "High"},
    source_modifier=activity(),
    target_modifier=activity("phos"),
)

sialic_acid_graph.add_directly_increases(
    cd33_phosphorylated,
    shp2,
    citation=citation,
    evidence=evidence_1,
    annotations={"Species": "9606", "Confidence": "High"},
    source_modifier=activity(),
    target_modifier=activity("phos"),
)

sialic_acid_graph.add_directly_decreases(
    shp1,
    syk,
    citation=citation,
    evidence=evidence_2,
    annotations={"Species": "9606", "Confidence": "High"},
    source_modifier=activity(),
    target_modifier=activity(),
)

sialic_acid_graph.add_directly_decreases(
    shp2,
    syk,
    citation=citation,
    evidence=evidence_2,
    annotations={"Species": "9606", "Confidence": "High"},
    source_modifier=activity(),
    target_modifier=activity(),
)

sialic_acid_graph.add_increases(
    syk,
    trem2,
    citation=citation,
    evidence=evidence_2,
    annotations={"Species": "9606", "Confidence": "Low"},
    source_modifier=activity(),
    target_modifier=activity(),
)

sialic_acid_graph.add_increases(
    syk,
    dap12,
    citation=citation,
    evidence=evidence_2,
    annotations={"Species": "9606", "Confidence": "Low"},
    source_modifier=activity(),
    target_modifier=activity(),
)



# -*- coding: utf-8 -*-

"""Tests for the repository class."""

import os
import tempfile

from pybel import to_bel_script, to_nodelink_file, to_pickle
from pybel.examples import egf_graph
from pybel.repository import BELRepository
from pybel.testing.cases import TemporaryCacheMixin


class TestRepository(TemporaryCacheMixin):
    """Tests for the repository class."""

    def test_repository(self):
        """Test the repository class."""
        name = "egf.bel"

        with tempfile.TemporaryDirectory() as temporary_directory:
            bel_path = os.path.join(temporary_directory, name)
            json_path = os.path.join(temporary_directory, f"{name}.json")
            pickle_path = os.path.join(temporary_directory, f"{name}.pickle")
            to_bel_script(egf_graph, bel_path)
            to_nodelink_file(egf_graph, json_path)
            to_pickle(egf_graph, pickle_path)

            repository = BELRepository(temporary_directory)
            graphs = repository.get_graphs(
                manager=self.manager,
                use_cached=True,
                use_tqdm=False,
            )
            self.assertNotEqual(0, len(graphs), msg="No graphs returned")
            self.assertEqual(1, len(graphs))
            self.assertIn(bel_path, graphs)
            graph = graphs[bel_path]
            self.assertEqual(graph.document, egf_graph.document)
            self.assertEqual(
                set(graph.nodes()),
                set(egf_graph.nodes()),
                msg=f"""
            Original nodes: {set(egf_graph.nodes())}
            New nodes:      {set(graph.nodes())}
            """,
            )
            self.assertEqual(
                set(graph.edges()),
                set(egf_graph.edges()),
                msg=f"""
            Original edges: {set(egf_graph.edges())}
            New edges:      {set(graph.edges())}
            """,
            )

            self.assertTrue(os.path.exists(json_path))
            self.assertTrue(os.path.exists(pickle_path))



# -*- coding: utf-8 -*-

"""Predicate functions for nodes based on their incident edges' relations."""

from typing import Set

from ...graph import BELGraph
from ....constants import CAUSAL_RELATIONS, RELATION
from ....dsl import BaseEntity

__all__ = [
    "has_in_edges",
    "has_causal_out_edges",
    "has_causal_in_edges",
    "has_causal_edges",
    "has_out_edges",
    "is_causal_central",
    "is_causal_sink",
    "is_causal_source",
    "no_causal_out_edges",
    "no_causal_in_edges",
    "no_out_edges",
    "no_in_edges",
    "no_causal_edges",
]


def has_in_edges(graph: BELGraph, node: BaseEntity, edge_types: Set[str]) -> bool:
    """Check if the node has any in-edges in the given set.

    :param graph: A BEL graph
    :param node: A BEL term
    :param edge_types: A collection of edge types to check against
    """
    return any(data[RELATION] in edge_types for _, _, data in graph.in_edges(node, data=True))


def no_in_edges(graph: BELGraph, node: BaseEntity, edge_types: Set[str]) -> bool:
    """Check if the node does not have any in-edges in the given set.

    :param graph: A BEL graph
    :param node: A BEL term
    :param edge_types: A collection of edge types to check against
    """
    return all(data[RELATION] not in edge_types for _, _, data in graph.in_edges(node, data=True))


def has_out_edges(graph: BELGraph, node: BaseEntity, edge_types: Set[str]) -> bool:
    """Check if the node has any out-edges in the given set.

    :param graph: A BEL graph
    :param node: A BEL term
    :param edge_types: A collection of edge types to check against
    """
    return any(data[RELATION] in edge_types for _, _, data in graph.out_edges(node, data=True))


def no_out_edges(graph: BELGraph, node: BaseEntity, edge_types: Set[str]) -> bool:
    """Check if the node does not have any out-edges in the given set.

    :param graph: A BEL graph
    :param node: A BEL term
    :param edge_types: A collection of edge types to check against
    """
    return all(data[RELATION] not in edge_types for _, _, data in graph.out_edges(node, data=True))


def has_causal_in_edges(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node has any causal in-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return has_in_edges(graph, node, CAUSAL_RELATIONS)


def no_causal_in_edges(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node does not have any causal in-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return no_in_edges(graph, node, CAUSAL_RELATIONS)


def has_causal_out_edges(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node has any causal out-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return has_out_edges(graph, node, CAUSAL_RELATIONS)


def no_causal_out_edges(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node does not have any causal out-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return no_out_edges(graph, node, CAUSAL_RELATIONS)


def has_causal_edges(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node has any causal out-edges or in-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return has_causal_in_edges(graph, node) or has_causal_out_edges(graph, node)


def no_causal_edges(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node does not have any causal out-edges or in-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return no_causal_in_edges(graph, node) and no_causal_out_edges(graph, node)


def is_causal_source(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node has causal out-edges but no causal in-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return no_causal_in_edges(graph, node) and has_causal_out_edges(graph, node)


def is_causal_sink(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node has causal in-edges but no causal out-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return has_causal_in_edges(graph, node) and no_causal_out_edges(graph, node)


def is_causal_central(graph: BELGraph, node: BaseEntity) -> bool:
    """Check if the node has both causal in-edges and also causal out-edges.

    :param graph: A BEL graph
    :param node: A BEL term
    """
    return has_causal_in_edges(graph, node) and has_causal_out_edges(graph, node)



# -*- coding: utf-8 -*-

"""Testing for CX and NDEx import/export."""

import os
import tempfile

from pybel import BELGraph, from_cx, from_cx_file, to_cx, to_cx_file
from pybel.examples import braf_graph, egf_graph, sialic_acid_graph, statin_graph
from tests.test_io.test_cx.cases import TestCase
from tests.test_io.test_cx.examples import example_graph


class TestSchema1(TestCase):
    """Test mapping schema 1."""

    def help_test_graph(self, graph: BELGraph) -> None:
        """Help test a graph round trip through a JSON object."""
        graph_cx = to_cx(graph)
        reconstituted = from_cx(graph_cx)
        self.assert_graph_equal(graph, reconstituted)

    def help_test_file(self, graph: BELGraph) -> None:
        """Help test a graph round trip through a file."""
        fd, path = tempfile.mkstemp()

        with open(path, "w") as file:
            to_cx_file(graph, file)

        with open(path) as file:
            reconstituted = from_cx_file(file)

        self.assert_graph_equal(graph, reconstituted)

        os.close(fd)
        os.remove(path)

    def test_sialic_acid_graph(self):
        """Test the round trip in the sialic acid graph."""
        self.help_test_graph(sialic_acid_graph)

    def test_braf_graph(self):
        """Test the round trip in the BRAF graph."""
        self.help_test_graph(braf_graph)

    def test_egf_graph(self):
        """Test the round trip in the EGF graph."""
        self.help_test_graph(egf_graph)

    def test_statin_graph(self):
        """Test the round trip in the statin graph."""
        self.help_test_graph(statin_graph)

    def test_example(self):
        """Test the round trip in an example graph."""
        self.help_test_graph(example_graph)

    def test_example_jsons(self):
        """Test the round trip to a JSON string with the example graph."""
        self.help_test_graph(example_graph)

    def test_example_file(self):
        """Test the round trip to a file with the example graph."""
        self.help_test_file(example_graph)
